from __future__ import annotations
import copy
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from valida.datapath import DataPath
from valida.rules import Rule

from .json_like import ChildObjectSpec, JSONLike
from .command_files import FileSpec, InputFile
from .element import ElementFilter, ElementGroup
from .errors import (
    ExtraInputs,
    MissingInputs,
    TaskTemplateInvalidNesting,
    TaskTemplateMultipleInputValues,
    TaskTemplateMultipleSchemaObjectives,
    TaskTemplateUnexpectedInput,
    TaskTemplateUnexpectedSequenceInput,
)
from .parameters import (
    InputSource,
    InputSourceType,
    InputValue,
    ParameterPath,
    SchemaInput,
    SchemaOutput,
    ValueSequence,
)
from .utils import (
    get_duplicate_items,
    get_in_container,
    get_item_repeat_index,
    get_relative_path,
    group_by_dict_key_values,
    set_in_container,
)


INPUT_SOURCE_TYPES = ["local", "default", "task", "import"]


class ElementSet(JSONLike):
    """Class to represent a parametrisation of a new set of elements."""

    _child_objects = (
        ChildObjectSpec(
            name="inputs",
            class_name="InputValue",
            is_multiple=True,
            dict_key_attr="parameter",
            dict_val_attr="value",
            parent_ref="_element_set",
        ),
        ChildObjectSpec(
            name="input_files",
            class_name="InputFile",
            is_multiple=True,
            parent_ref="_element_set",
        ),
        ChildObjectSpec(
            name="resources",
            class_name="ResourceList",
            parent_ref="_element_set",
        ),
        ChildObjectSpec(
            name="sequences",
            class_name="ValueSequence",
            is_multiple=True,
            parent_ref="_element_set",
        ),
        ChildObjectSpec(
            name="input_sources",
            class_name="InputSource",
            is_multiple=True,
            is_dict_values=True,
            is_dict_values_ensure_list=True,
        ),
    )

    def __init__(
        self,
        inputs: Optional[List[InputValue]] = None,
        input_files: Optional[List[InputFile]] = None,
        sequences: Optional[List[ValueSequence]] = None,
        resources: Optional[Dict[str, Dict]] = None,
        repeats: Optional[Union[int, List[int]]] = 1,
        input_sources: Optional[Dict[str, InputSource]] = None,
        nesting_order: Optional[List] = None,
        sourceable_elements: Optional[List[int]] = None,
        allow_non_coincident_task_sources: Optional[bool] = False,
    ):
        """
        Parameters
        ----------
        sourceable_elements
            If specified, a list of global element indices from which inputs for the new
            elements associated with this element set may be sourced. If not specified,
            all workflow elements are considered sourceable.
        allow_non_coincident_task_sources
            If True, if more than one parameter is sourced from the same task, then allow
            these sources to come from distinct element sub-sets. If False (default),
            only the intersection of element sub-sets for all parameters are included.

        """

        if isinstance(resources, dict):
            resources = self.app.ResourceList.from_json_like(resources)
        elif isinstance(resources, list):
            resources = self.app.ResourceList(resources)
        elif not resources:
            resources = self.app.ResourceList([self.app.ResourceSpec()])

        self.inputs = inputs or []
        self.input_files = input_files or []
        self.repeats = repeats
        self.resources = resources
        self.sequences = sequences or []
        self.input_sources = input_sources or {}
        self.nesting_order = nesting_order or {}
        self.sourceable_elements = sourceable_elements
        self.allow_non_coincident_task_sources = allow_non_coincident_task_sources

        self._validate()
        self._set_parent_refs()

        self._task_template = None  # assigned by parent Task
        self._defined_input_types = None  # assigned on _task_template assignment
        self._element_local_idx_range = None  # assigned by WorkflowTask._add_element_set
        self._element_global_idx_range = None  # assigned by WorkflowTask._add_element_set

    def __deepcopy__(self, memo):
        dct = self.to_dict()
        elem_local_idx_range = dct.pop("_element_local_idx_range", None)
        elem_global_idx_range = dct.pop("_element_global_idx_range", None)
        obj = self.__class__(**copy.deepcopy(dct, memo))
        obj._task_template = self._task_template
        obj._defined_input_types = self._defined_input_types
        obj._element_local_idx_range = elem_local_idx_range
        obj._element_global_idx_range = elem_global_idx_range
        return obj

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.to_dict() == other.to_dict():
            return True
        return False

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        orig_inp = json_like.pop("original_input_sources", None)
        orig_nest = json_like.pop("original_nesting_order", None)
        elem_local_idx_range = json_like.pop("_element_local_idx_range", None)
        elem_global_idx_range = json_like.pop("_element_global_idx_range", None)
        obj = cls(**json_like)
        obj.original_input_sources = orig_inp
        obj.original_nesting_order = orig_nest
        obj._element_local_idx_range = elem_local_idx_range
        obj._element_global_idx_range = elem_global_idx_range
        return obj

    def prepare_persistent_copy(self):
        """Return a copy of self, which will then be made persistent, and save copies of
        attributes that may be changed during integration with the workflow."""
        obj = copy.deepcopy(self)
        obj.original_nesting_order = self.nesting_order
        obj.original_input_sources = self.input_sources
        return obj

    def to_dict(self):
        dct = super().to_dict()
        del dct["_defined_input_types"]
        del dct["_task_template"]
        return dct

    @property
    def task_template(self):
        return self._task_template

    @task_template.setter
    def task_template(self, value):
        self._task_template = value
        self._validate_against_template()

    @property
    def input_types(self):
        return [i.parameter.typ for i in self.inputs]

    @property
    def element_local_idx_range(self):
        return tuple(self._element_local_idx_range)

    @property
    def element_global_idx_range(self):
        return tuple(self._element_global_idx_range)

    def _validate(self):
        dup_params = get_duplicate_items(self.input_types)
        if dup_params:
            raise TaskTemplateMultipleInputValues(
                f"The following parameters are associated with multiple input value "
                f"definitions: {dup_params!r}."
            )

    def _validate_against_template(self):

        unexpected_types = (
            set(self.input_types) - self.task_template.all_schema_input_types
        )
        if unexpected_types:
            raise TaskTemplateUnexpectedInput(
                f"The following input parameters are unexpected: {list(unexpected_types)!r}"
            )

        seq_inp_types = []
        for seq_i in self.sequences:
            inp_type = seq_i.input_type
            if inp_type:
                bad_inp = {inp_type} - self.task_template.all_schema_input_types
                allowed_str = ", ".join(
                    f'"{i}"' for i in self.task_template.all_schema_input_types
                )
                if bad_inp:
                    raise TaskTemplateUnexpectedSequenceInput(
                        f"The input type {inp_type!r} specified in the following sequence"
                        f" path is unexpected: {seq_i.path!r}. Available input types are: "
                        f"{allowed_str}."
                    )
                seq_inp_types.append(inp_type)
            if seq_i.path not in self.nesting_order:
                self.nesting_order.update({seq_i.path: seq_i.nesting_order})

        for k, v in self.nesting_order.items():
            if v < 0:
                raise TaskTemplateInvalidNesting(
                    f"`nesting_order` must be >=0 for all keys, but for key {k!r}, value "
                    f"of {v!r} was specified."
                )

        self._defined_input_types = set(self.input_types + seq_inp_types)

    @classmethod
    def ensure_element_sets(
        cls,
        inputs=None,
        input_files=None,
        sequences=None,
        resources=None,
        repeats=None,
        input_sources=None,
        nesting_order=None,
        element_sets=None,
        sourceable_elements=None,
    ):
        args = (
            inputs,
            input_files,
            sequences,
            resources,
            repeats,
            input_sources,
            nesting_order,
        )
        args_not_none = [i is not None for i in args]

        if any(args_not_none):
            if element_sets is not None:
                raise ValueError(
                    "If providing an `element_set`, no other arguments are allowed."
                )
            else:
                element_sets = [cls(*args, sourceable_elements=sourceable_elements)]
        else:
            if element_sets is None:
                element_sets = [cls(*args, sourceable_elements=sourceable_elements)]

        return element_sets

    @property
    def defined_input_types(self):
        return self._defined_input_types

    @property
    def undefined_input_types(self):
        return self.task_template.all_schema_input_types - self.defined_input_types

    def get_sequence_from_path(self, sequence_path):
        for i in self.sequences:
            if i.path == sequence_path:
                return i

    def get_defined_parameter_types(self):
        out = []
        for inp in self.inputs:
            if not inp.is_sub_value:
                out.append(inp.normalised_inputs_path)
        for seq in self.sequences:
            if seq.parameter and not seq.is_sub_value:  # ignore resource sequences
                out.append(seq.normalised_inputs_path)
        return out

    def get_defined_sub_parameter_types(self):
        out = []
        for inp in self.inputs:
            if inp.is_sub_value:
                out.append(inp.normalised_inputs_path)
        for seq in self.sequences:
            if seq.parameter and seq.is_sub_value:  # ignore resource sequences
                out.append(seq.normalised_inputs_path)
        return out

    def get_locally_defined_inputs(self):
        return self.get_defined_parameter_types() + self.get_defined_sub_parameter_types()

    def get_sequence_by_path(self, path):
        for seq in self.sequences:
            if seq.path == path:
                return seq

    @property
    def index(self):
        for idx, element_set in enumerate(self.task_template.element_sets):
            if element_set is self:
                return idx

    @property
    def task(self):
        return self.task_template.workflow_template.workflow.tasks[
            self.task_template.index
        ]

    @property
    def elements(self):
        elements = self.task.get_elements_of_element_set(self.index)
        return elements

    def get_task_dependencies(self, as_objects=False):
        """Get upstream tasks that this element set depends on."""
        deps = []
        for element in self.elements:
            for dep_i in element.get_task_dependencies(as_objects=False):
                if dep_i not in deps:
                    deps.append(dep_i)
        deps = sorted(deps)
        if as_objects:
            deps = [self.task.workflow.tasks.get(insert_ID=i) for i in deps]

        return deps


class Task(JSONLike):
    """Parametrisation of an isolated task for which a subset of input values are given
    "locally". The remaining input values are expected to be satisfied by other
    tasks/imports in the workflow."""

    _child_objects = (
        ChildObjectSpec(
            name="schemas",
            class_name="TaskSchema",
            is_multiple=True,
            shared_data_name="task_schemas",
            shared_data_primary_key="name",
            parent_ref="_task_template",
        ),
        ChildObjectSpec(
            name="element_sets",
            class_name="ElementSet",
            is_multiple=True,
            parent_ref="task_template",
        ),
    )

    def __init__(
        self,
        schemas: Union[TaskSchema, str, List[TaskSchema], List[str]],
        repeats: Optional[Union[int, List[int]]] = None,
        resources: Optional[Dict[str, Dict]] = None,
        inputs: Optional[List[InputValue]] = None,
        input_files: Optional[List[InputFile]] = None,
        sequences: Optional[List[ValueSequence]] = None,
        input_sources: Optional[Dict[str, InputSource]] = None,
        nesting_order: Optional[List] = None,
        element_sets: Optional[List[ElementSet]] = None,
        sourceable_elements: Optional[List[int]] = None,
    ):

        """
        Parameters
        ----------
        schema
            A (list of) `TaskSchema` object(s) and/or a (list of) strings that are task
            schema names that uniquely identify a task schema. If strings are provided,
            the `TaskSchema` object will be fetched from the known task schemas loaded by
            the app configuration.

        """

        # TODO: allow init via specifying objective and/or method and/or implementation
        # (lists of) strs e.g.: Task(
        #   objective='simulate_VE_loading',
        #   method=['CP_FFT', 'taylor'],
        #   implementation=['damask', 'damask']
        # )
        # where method and impl must be single strings of lists of the same length
        # and method/impl are optional/required only if necessary to disambiguate
        #
        # this would be like Task(schemas=[
        #   'simulate_VE_loading_CP_FFT_damask',
        #   'simulate_VE_loading_taylor_damask'
        # ])

        if not isinstance(schemas, list):
            schemas = [schemas]

        _schemas = []
        for i in schemas:
            if isinstance(i, str):
                try:
                    i = self.app.TaskSchema.get_by_key(i)
                except KeyError:
                    raise KeyError(f"TaskSchema {i!r} not found.")
            elif not isinstance(i, self.app.TaskSchema):
                raise TypeError(f"Not a TaskSchema object: {i!r}")
            _schemas.append(i)

        self._schemas = _schemas

        self._element_sets = self.app.ElementSet.ensure_element_sets(
            inputs=inputs,
            input_files=input_files,
            sequences=sequences,
            resources=resources,
            repeats=repeats,
            input_sources=input_sources,
            nesting_order=nesting_order,
            element_sets=element_sets,
            sourceable_elements=sourceable_elements,
        )

        self._validate()
        self._name = self._get_name()

        self.workflow_template = None  # assigned by parent WorkflowTemplate
        self._insert_ID = None
        self._dir_name = None

        self._set_parent_refs()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.to_dict() == other.to_dict():
            return True
        return False

    def _add_element_set(self, element_set: ElementSet):
        """Invoked by WorkflowTask._add_element_set."""
        self._element_sets.append(element_set)
        self.workflow_template.workflow._store.add_element_set(
            self.index, element_set.to_json_like()[0]
        )

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        insert_ID = json_like.pop("insert_ID", None)
        dir_name = json_like.pop("dir_name", None)
        obj = cls(**json_like)
        obj._insert_ID = insert_ID
        obj._dir_name = dir_name
        return obj

    def __repr__(self):
        return f"{self.__class__.__name__}(" f"name={self.name!r}" f")"

    def __deepcopy__(self, memo):
        kwargs = self.to_dict()
        _insert_ID = kwargs.pop("insert_ID")
        _dir_name = kwargs.pop("dir_name")
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._insert_ID = _insert_ID
        obj._dir_name = _dir_name
        obj._name = self._name
        obj.workflow_template = self.workflow_template
        return obj

    def to_persistent(self, workflow, insert_ID):
        """Return a copy where any schema input defaults are saved to a persistent
        workflow. Element set data is not made persistent."""

        obj = copy.deepcopy(self)
        new_refs = []
        source = {"type": "default_input", "task_insert_ID": insert_ID}
        for schema in obj.schemas:
            new_refs.extend(schema.make_persistent(workflow, source))

        return obj, new_refs

    def to_dict(self):
        out = super().to_dict()
        return {k.lstrip("_"): v for k, v in out.items() if k != "_name"}

    def set_sequence_parameters(self, element_set):
        # set ValueSequence Parameter objects:
        for seq in element_set.sequences:
            if seq.input_type:
                for schema_i in self.schemas:
                    for inp_j in schema_i.inputs:
                        if inp_j.typ == seq.input_type:
                            seq._parameter = inp_j.parameter

    def _validate(self):

        # TODO: check a nesting order specified for each sequence?

        names = set(i.objective.name for i in self.schemas)
        if len(names) > 1:
            raise TaskTemplateMultipleSchemaObjectives(
                f"All task schemas used within a task must have the same "
                f"objective, but found multiple objectives: {list(names)!r}"
            )

    def _get_name(self):
        out = f"{self.objective.name}"
        for idx, schema_i in enumerate(self.schemas, start=1):
            need_and = idx < len(self.schemas) and (
                self.schemas[idx].method or self.schemas[idx].implementation
            )
            out += (
                f"{f'_{schema_i.method}' if schema_i.method else ''}"
                f"{f'_{schema_i.implementation}' if schema_i.implementation else ''}"
                f"{f'_and' if need_and else ''}"
            )
        return out

    @staticmethod
    def get_task_unique_names(tasks: List[Task]):
        """Get the unique name of each in a list of tasks.

        Returns
        -------
        list of str

        """

        task_name_rep_idx = get_item_repeat_index(
            tasks,
            item_callable=lambda x: x.name,
            distinguish_singular=True,
        )

        names = []
        for idx, task in enumerate(tasks):
            add_rep = f"_{task_name_rep_idx[idx]}" if task_name_rep_idx[idx] > 0 else ""
            names.append(f"{task.name}{add_rep}")

        return names

    def _get_nesting_order(self, seq):
        """Find the nesting order for a task sequence."""
        return self.nesting_order[seq.normalised_path] if len(seq.values) > 1 else -1

    def _prepare_persistent_outputs(self, workflow, local_element_idx_range):
        # TODO: check that schema is present when adding task? (should this be here?)
        output_data_indices = {}
        for schema in self.schemas:
            for output in schema.outputs:

                # TODO: consider multiple schemas in action index?

                # Find the last action where specified output type is an output:
                output_act_idx = None
                for act_idx, act in enumerate(schema.actions):
                    if output.typ in act.get_output_types():
                        output_act_idx = act_idx

                if output_act_idx is None:
                    raise RuntimeError(
                        f"Output {output} does not appear in any schema actions."
                    )

                path = f"outputs.{output.typ}"
                output_data_indices[path] = []
                for idx in range(*local_element_idx_range):
                    param_src = {
                        "type": "EAR_output",
                        "task_insert_ID": self.insert_ID,
                        "element_idx": idx,
                        "action_idx": output_act_idx,
                        "run_idx": 0,
                    }
                    data_ref = workflow._add_unset_parameter_data(param_src)
                    output_data_indices[path].append(data_ref)

        return output_data_indices

    def prepare_element_resolution(self, element_set, input_data_indices):

        multiplicities = []
        for path_i, inp_idx_i in input_data_indices.items():
            multiplicities.append(
                {
                    "multiplicity": len(inp_idx_i),
                    "nesting_order": element_set.nesting_order.get(path_i, -1),
                    "path": path_i,
                }
            )

        return multiplicities

    @property
    def index(self):
        if self.workflow_template:
            return self.workflow_template.tasks.index(self)
        else:
            return None

    @property
    def _element_indices(self):
        return self.workflow_template.workflow.tasks[self.index].element_indices

    def get_available_task_input_sources(
        self,
        element_set: ElementSet,
        source_tasks: Optional[List[Task]] = None,
    ) -> List[InputSource]:
        """For each input parameter of this task, generate a list of possible input sources
        that derive from inputs or outputs of this and other provided tasks.

        Note this only produces a subset of available input sources for each input
        parameter; other available input sources may exist from workflow imports."""

        # TODO: also search sub-parameters in the source tasks!

        available = {}
        for inputs_path, inp_info in self.get_all_inputs_info(element_set).items():

            available[inputs_path] = []

            # local specification takes precedence:
            if inputs_path in element_set.get_locally_defined_inputs():
                available[inputs_path].append(self.app.InputSource.local())

            # search for task sources:
            for src_task_i in source_tasks or []:

                for param_i in src_task_i.provides_parameters:

                    if param_i.typ == inputs_path:

                        if param_i.input_or_output == "input":
                            # input parameter might not be provided e.g. if it only used
                            # to generate an input file, and that input file is passed
                            # directly, so consider only source task element sets that
                            # provide the input:
                            es_idx = src_task_i.get_param_provided_element_sets(
                                param_i.typ
                            )
                        else:
                            # outputs are always available, so consider all source task
                            # element sets:
                            es_idx = range(src_task_i.num_element_sets)

                        if not es_idx:
                            continue
                        else:
                            src_elems = []
                            for es_idx_i in es_idx:
                                es_i = src_task_i.element_sets[es_idx_i]
                                src_elems += list(range(*es_i.element_global_idx_range))

                        if element_set.sourceable_elements is not None:
                            # can only use a subset of elements (this is the case where
                            # this element set is generated from an upstream element set,
                            # in which case we only want to consider newly added upstream
                            # elements when adding elements from this element set):
                            src_elems = list(
                                set(element_set.sourceable_elements) & set(src_elems)
                            )
                            if not src_elems:
                                continue

                        task_source = self.app.InputSource.task(
                            task_ref=src_task_i.insert_ID,
                            task_source_type=param_i.input_or_output,
                            elements=src_elems,
                        )
                        available[inputs_path].append(task_source)

            if inp_info["has_default"]:
                available[inputs_path].append(self.app.InputSource.default())

        return available

    @property
    def schemas(self):
        return self._schemas

    @property
    def element_sets(self):
        return self._element_sets

    @property
    def num_element_sets(self):
        return len(self._element_sets)

    @property
    def insert_ID(self):
        return self._insert_ID

    @property
    def dir_name(self):
        "Artefact directory name."
        return self._dir_name

    @property
    def name(self):
        return self._name

    @property
    def objective(self):
        return self.schemas[0].objective

    @property
    def all_schema_inputs(self) -> Tuple[SchemaInput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.inputs)

    @property
    def all_schema_outputs(self) -> Tuple[SchemaOutput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.outputs)

    @property
    def all_schema_input_types(self):
        """Get the set of all schema input types (over all specified schemas)."""
        return {inp_j for schema_i in self.schemas for inp_j in schema_i.input_types}

    @property
    def all_schema_input_normalised_paths(self):
        return {f"inputs.{i}" for i in self.all_schema_input_types}

    @property
    def all_schema_output_types(self):
        """Get the set of all schema output types (over all specified schemas)."""
        return {out_j for schema_i in self.schemas for out_j in schema_i.output_types}

    def get_schema_action(self, idx):
        _idx = 0
        for schema in self.schemas:
            for action in schema.actions:
                if _idx == idx:
                    return action
                _idx += 1
        raise ValueError(f"No action in task {self.name!r} with index {idx!r}.")

    def all_schema_actions(self) -> Iterator[Tuple[int, Action]]:
        idx = 0
        for schema in self.schemas:
            for action in schema.actions:
                yield (idx, action)
                idx += 1

    @property
    def all_sourced_normalised_paths(self):
        sourced_input_types = []
        for elem_set in self.element_sets:
            for inp in elem_set.inputs:
                if inp.is_sub_value:
                    sourced_input_types.append(inp.normalised_path)
            for seq in elem_set.sequences:
                if seq.is_sub_value:
                    sourced_input_types.append(seq.normalised_path)
        return set(sourced_input_types) | self.all_schema_input_normalised_paths

    def is_input_type_required(self, typ, element_set):

        provided_files = [i.file for i in element_set.input_files]
        # required if is appears in any command:
        for schema in self.schemas:
            for act in schema.actions:
                if typ in act.get_command_input_types():
                    return True

                # required if used in any input file generators and input file is not
                # provided:
                for IFG in act.input_file_generators:
                    if typ in (i.typ for i in IFG.inputs):
                        if IFG.input_file not in provided_files:
                            return True

        return False

    def is_input_type_provided(self, typ: str, element_set: ElementSet) -> bool:
        """Check if an input is provided as an InputValue or a ValueSequence."""
        for inp in element_set.inputs:
            if typ == inp.parameter.typ:
                return True

        for seq in element_set.sequences:
            if not seq.is_sub_value and typ == seq.parameter.typ:
                return True

        return False

    def get_param_provided_element_sets(self, typ: str) -> List[int]:
        """Get the element set indices of this task for which a specified parameter type
        is provided."""
        es_idx = []
        for idx, src_es in enumerate(self.element_sets):
            if self.is_input_type_provided(typ, src_es):
                es_idx.append(idx)
        return es_idx

    def get_all_inputs_info(self, element_set):
        """Get a dict whose keys are the normalised paths (without the "inputs" prefix),
        and whose values are the associated default value InputValue object, in the case
        the input is a SchemaInput, and a default is defined.

        # TODO update docstring

        Parameters
        ----------
        element_set : ElementSet
            Find inputs and sequences in this element set that have sub-parameter paths.

        """

        info = {}
        for schema_input in self.all_schema_inputs:
            info[schema_input.parameter.typ] = {
                "has_default": schema_input.default_value is not None
            }

        for inp_path in element_set.get_defined_sub_parameter_types():
            info[inp_path] = {"has_default": False}

        for inp in info:
            info[inp]["is_required"] = self.is_input_type_required(inp, element_set)
            info[inp]["is_provided"] = self.is_input_type_provided(inp, element_set)

        return info

    def get_all_required_schema_inputs(self, element_set):
        info = self.get_all_inputs_info(element_set)
        return tuple(
            i for i in self.all_schema_inputs if info[i.parameter.typ]["is_required"]
        )

    @property
    def all_sequences_normalised_paths(self):
        return [j.normalised_path for i in self.element_sets for j in i.sequences]

    @property
    def all_used_sequences_normalised_paths(self):
        return [
            j.normalised_path
            for i in self.element_sets
            for j in i.sequences
            if not j.is_unused
        ]

    @property
    def universal_input_types(self):
        """Get input types that are associated with all schemas"""

    @property
    def non_universal_input_types(self):
        """Get input types for each schema that are non-universal."""

    @property
    def defined_input_types(self):
        return self._defined_input_types

    @property
    def undefined_input_types(self):
        return self.all_schema_input_types - self.defined_input_types

    @property
    def undefined_inputs(self):
        return [
            inp_j
            for schema_i in self.schemas
            for inp_j in schema_i.inputs
            if inp_j.typ in self.undefined_input_types
        ]

    @property
    def unsourced_inputs(self):
        """Get schema input types for which no input sources are currently specified."""
        return self.all_schema_input_types - set(self.input_sources.keys())

    @property
    def provides_parameters(self):
        return tuple(j for schema in self.schemas for j in schema.provides_parameters)

    @property
    def _metadata(self):
        return self.workflow_template.workflow.metadata["template"]["tasks"][self.index]

    def get_sub_parameter_input_values(self):
        return [i for i in self.inputs if i.is_sub_value]

    def get_non_sub_parameter_input_values(self):
        return [i for i in self.inputs if not i.is_sub_value]

    def add_group(
        self, name: str, where: ElementFilter, group_by_distinct: ParameterPath
    ):
        group = ElementGroup(name=name, where=where, group_by_distinct=group_by_distinct)
        self.groups.add_object(group)


class WorkflowTask:
    """Class to represent a Task that is bound to a Workflow."""

    _app_attr = "app"

    def __init__(
        self,
        workflow: Workflow,
        template: Task,
        index: int,
        num_elements: int,
    ):
        self._workflow = workflow
        self._template = template
        self._index = index
        self._num_elements = num_elements

        # assigned/incremented when new elements are added and reset on dump to disk:
        self._pending_num_elements = 0

        self._elements = None  # assigned on `elements` first access

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.unique_name!r})"

    def _reset_pending_elements(self):
        self._pending_num_elements = 0

    def _accept_pending_elements(self):
        self._num_elements = self.num_elements
        self._reset_pending_elements()

    @classmethod
    def new_empty_task(cls, workflow: Workflow, template: Task, index: int):
        obj = cls(
            workflow=workflow,
            template=template,
            index=index,
            num_elements=0,
        )
        return obj

    @property
    def workflow(self):
        return self._workflow

    @property
    def template(self):
        return self._template

    @property
    def index(self):
        return self._index

    @property
    def num_elements(self):
        return self._num_elements + self._pending_num_elements

    @property
    def name(self):
        return self.template.name

    @property
    def unique_name(self):
        return self.workflow.get_task_unique_names()[self.index]

    @property
    def insert_ID(self):
        return self.template.insert_ID

    @property
    def dir_name(self):
        return self.template.dir_name

    @property
    def num_element_sets(self):
        return self.template.num_element_sets

    @property
    def elements(self):
        if self._elements is None:
            self._elements = self.app.Elements(self)
        return self._elements

    @property
    def dir_path(self):
        return self.workflow.path / "tasks" / self.dir_name

    @property
    def element_dir_list_file_path(self):
        return self.dir_path / "element_dirs.txt"

    @property
    def run_script_file_path(self):
        return self.dir_path / "run_script.ps1"

    @property
    def _metadata(self):
        return self.workflow.metadata["tasks"][self.index]

    def write_element_dirs(self):
        self.dir_path.mkdir(exist_ok=True, parents=True)
        elem_paths = [self.dir_path / elem.dir_name for elem in self.elements]
        for path_i in elem_paths:
            path_i.mkdir(exist_ok=True)

        # write a text file whose lines correspond to element paths
        with self.element_dir_list_file_path.open("wt") as fp:
            for elem in elem_paths:
                fp.write(f"{elem}\n")

    def _make_new_elements_persistent(self, element_set, element_set_idx):
        """Save parameter data to the persistent workflow."""

        input_data_idx = {}
        sequence_idx = {}

        # Assign first assuming all locally defined values are to be used:
        param_src = {
            "type": "local_input",
            "task_insert_ID": self.insert_ID,
            "element_set_idx": element_set_idx,
        }
        for res_i in element_set.resources:
            key, dat_ref, _ = res_i.make_persistent(self.workflow, param_src)
            input_data_idx[key] = dat_ref

        for inp_i in element_set.inputs:
            key, dat_ref, _ = inp_i.make_persistent(self.workflow, param_src)
            input_data_idx[key] = dat_ref

        for inp_file_i in element_set.input_files:
            key, dat_ref, _ = inp_file_i.make_persistent(self.workflow, param_src)
            input_data_idx[key] = dat_ref

        for seq_i in element_set.sequences:
            key, dat_ref, _ = seq_i.make_persistent(self.workflow, param_src)
            input_data_idx[key] = dat_ref
            sequence_idx[key] = list(range(len(dat_ref)))

        # Now check for task- and default-sources and overwrite or append to local sources:
        for schema_input in self.template.get_all_required_schema_inputs(element_set):

            key = f"inputs.{schema_input.typ}"
            sources = element_set.input_sources[schema_input.typ]

            for inp_src in sources:

                if inp_src.source_type is InputSourceType.TASK:

                    src_task = inp_src.get_task(self.workflow)

                    src_elements = [i for i in src_task.elements]  # TODO: ??
                    if inp_src.elements:
                        # only include "sourceable" elements:
                        src_elements = [
                            i for i in src_elements if i.global_idx in inp_src.elements
                        ]

                    if not src_elements:
                        continue

                    task_source_type = inp_src.task_source_type.name.lower()
                    src_key = f"{task_source_type}s.{schema_input.typ}"
                    grp_idx = [elem.get_data_idx()[src_key] for elem in src_elements]

                    if self.app.InputSource.local() in sources:
                        # add task source to existing local source:
                        input_data_idx[key] += grp_idx

                    else:
                        # overwrite existing local source (if it exists):
                        input_data_idx[key] = grp_idx
                        if key in sequence_idx:
                            sequence_idx.pop(key)
                            seq = element_set.get_sequence_by_path(key)
                            seq.is_unused = True

                if inp_src.source_type is InputSourceType.DEFAULT:

                    grp_idx = [schema_input.default_value._value_group_idx]
                    if self.app.InputSource.local() in sources:
                        input_data_idx[key] += grp_idx

                    else:
                        input_data_idx[key] = grp_idx

        # sort smallest to largest path, so more-specific items overwrite less-specific
        # items parameter retrieval:
        # TODO: is this still necessary?
        # data_idx_paths = sorted(input_data_idx.keys(), key=lambda x: len(x.split(".")))
        # input_data_idx = {data_idx_paths.index(k): v for k, v in input_data_idx.items()}
        # input_sources = {data_idx_paths.index(k): v for k, v in input_sources.items()}

        return (input_data_idx, sequence_idx)

    def ensure_input_sources(self, element_set):
        """Check valid input sources are specified for a new task to be added to the
        workflow in a given position. If none are specified, set them according to the
        default behaviour."""

        # this just depends on this schema and other schemas:
        available_sources = self.template.get_available_task_input_sources(
            element_set=element_set,
            source_tasks=self.workflow.template.tasks[: self.index],
        )  # TODO: test all parameters have a key here?

        # TODO: get available input sources from workflow imports

        all_inputs_info = self.template.get_all_inputs_info(element_set)

        # check any specified sources are valid:
        for inputs_path in all_inputs_info:
            for specified_source in element_set.input_sources.get(inputs_path, []):
                self.workflow._resolve_input_source_task_reference(
                    specified_source, self.unique_name
                )
                if not specified_source.is_in(available_sources[inputs_path]):
                    raise ValueError(
                        f"The input source {specified_source.to_string()!r} is not "
                        f"available for input path {inputs_path!r}. Available "
                        f"input sources are: "
                        f"{[i.to_string() for i in available_sources[inputs_path]]}"
                    )

        # an input is not required if it is only used to generate an input file that is
        # passed directly:
        req_types = set(k for k, v in all_inputs_info.items() if v["is_required"])
        unsourced_inputs = req_types - set(element_set.input_sources.keys())

        extra_types = set(
            k
            for k, v in all_inputs_info.items()
            if not v["is_required"] and v["is_provided"]
        )
        if extra_types:
            extra_str = ", ".join(f"{i!r}" for i in extra_types)
            raise ExtraInputs(
                message=(
                    f"The following inputs are not required, but have been passed: "
                    f"{extra_str}."
                ),
                extra_inputs=extra_types,
            )

        # set source for any unsourced inputs:
        missing = []
        for input_type in unsourced_inputs:
            inp_i_sources = available_sources[input_type]
            source = None
            try:
                # first element is defined by default to take precedence in
                # `get_available_task_input_sources`:
                source = inp_i_sources[0]
            except IndexError:
                missing.append(input_type)

            if source is not None:
                element_set.input_sources.update({input_type: [source]})

        # TODO: collate all input sources separately, then can fall back to a different
        # input source (if it was not specified manually) and if the "top" input source
        # results in no available elements due to `allow_non_coincident_task_sources`.

        if not element_set.allow_non_coincident_task_sources:

            sources_by_task = {}
            for inp_type, sources in element_set.input_sources.items():
                source = sources[0]
                if source.source_type is InputSourceType.TASK:
                    if source.task_ref not in sources_by_task:
                        sources_by_task[source.task_ref] = {}
                    sources_by_task[source.task_ref][inp_type] = source

            # if multiple parameters are sourced from the same upstream task, only use
            # elements for which all parameters are available (the set intersection):
            for sources in sources_by_task.values():
                first_src = next(iter(sources.values()))
                intersect_task_i = set(first_src.elements)
                for src_i in sources.values():
                    intersect_task_i.intersection_update(src_i.elements)

                # now change elements for the affected input sources:
                for inp_type in sources.keys():
                    element_set.input_sources[inp_type][0].elements = list(
                        intersect_task_i
                    )

        if missing:
            missing_str = ", ".join(f"{i!r}" for i in missing)
            raise MissingInputs(
                message=f"The following inputs have no sources: {missing_str}.",
                missing_inputs=missing,
            )

    def generate_new_elements(
        self,
        input_data_indices,
        output_data_indices,
        element_data_indices,
        sequence_indices,
    ):

        new_elements = []
        element_sequence_indices = {}
        for i_idx, i in enumerate(element_data_indices):
            elem_i = {k: input_data_indices[k][v] for k, v in i.items()}
            elem_i.update({k: v[i_idx] for k, v in output_data_indices.items()})
            new_elements.append(elem_i)

            # track which sequence value indices (if any) are used for each new element:
            for k, v in i.items():
                if k in sequence_indices:
                    if k not in element_sequence_indices:
                        element_sequence_indices[k] = []
                    element_sequence_indices[k].append(sequence_indices[k][v])

        return new_elements, element_sequence_indices

    @property
    def upstream_tasks(self):
        return [task for task in self.workflow.tasks[: self.index]]

    @property
    def downstream_tasks(self):
        return [task for task in self.workflow.tasks[self.index + 1 :]]

    def get_elements_of_element_set(self, element_set_index):
        elem_slice = slice(
            *self.template.element_sets[element_set_index].element_local_idx_range
        )
        return self.elements[elem_slice]

    @staticmethod
    def resolve_element_data_indices(multiplicities):
        """Find the index of the Zarr parameter group index list corresponding to each
        input data for all elements.

        Parameters
        ----------
        multiplicities : list of dict
            Each list item represents a sequence of values with keys:
                multiplicity: int
                nesting_order: int
                path : str

        Returns
        -------
        element_dat_idx : list of dict
            Each list item is a dict representing a single task element and whose keys are
            input data paths and whose values are indices that index the values of the
            dict returned by the `task.make_persistent` method.

        """

        # order by nesting order (so lower nesting orders will be fastest-varying):
        multi_srt = sorted(multiplicities, key=lambda x: x["nesting_order"])
        multi_srt_grp = group_by_dict_key_values(
            multi_srt, "nesting_order"
        )  # TODO: is tested?

        element_dat_idx = [{}]
        for para_sequences in multi_srt_grp:

            # check all equivalent nesting_orders have equivalent multiplicities
            all_multis = {i["multiplicity"] for i in para_sequences}
            if len(all_multis) > 1:
                raise ValueError(
                    f"All inputs with the same `nesting_order` must have the same "
                    f"multiplicity, but for paths "
                    f"{[i['path'] for i in para_sequences]} with "
                    f"`nesting_order` {para_sequences[0]['nesting_order']} found "
                    f"multiplicities {[i['multiplicity'] for i in para_sequences]}."
                )

            new_elements = []
            for val_idx in range(para_sequences[0]["multiplicity"]):
                for element in element_dat_idx:
                    new_elements.append(
                        {
                            **element,
                            **{i["path"]: val_idx for i in para_sequences},
                        }
                    )
            element_dat_idx = new_elements

        return element_dat_idx

    def _add_element_set(self, element_set):
        """
        Returns
        -------
        element_indices : list of int
            Global indices of newly added elements.

        """

        self.template.set_sequence_parameters(element_set)

        self.ensure_input_sources(element_set)  # may modify element_set.input_sources

        (input_data_idx, seq_idx) = self._make_new_elements_persistent(
            element_set=element_set,
            element_set_idx=self.num_element_sets,
        )

        element_set.task_template = self.template  # may modify element_set.nesting_order

        multiplicities = self.template.prepare_element_resolution(
            element_set, input_data_idx
        )

        element_inp_data_idx = self.resolve_element_data_indices(multiplicities)

        global_element_idx_range = [
            self.workflow.num_elements,
            self.workflow.num_elements + len(element_inp_data_idx),
        ]
        local_element_idx_range = [
            self.num_elements,
            self.num_elements + len(element_inp_data_idx),
        ]

        # associate global element indices (as a range) with the element set:
        element_set._element_global_idx_range = global_element_idx_range
        element_set._element_local_idx_range = local_element_idx_range
        self.template._add_element_set(element_set)

        output_data_idx = self.template._prepare_persistent_outputs(
            workflow=self.workflow,
            local_element_idx_range=local_element_idx_range,
        )

        (element_data_idx, element_seq_idx) = self.generate_new_elements(
            input_data_idx,
            output_data_idx,
            element_inp_data_idx,
            seq_idx,
        )

        elements = []

        for elem_idx, data_idx in enumerate(element_data_idx):
            schema_params = set(i for i in data_idx.keys() if len(i.split(".")) == 2)
            data_idx_i = data_idx
            action_runs = {}
            for act_idx, action in self.template.all_schema_actions():
                if all(self.test_action_rule(i, data_idx) for i in action.rules):

                    param_source = {
                        "type": "EAR_output",
                        "task_insert_ID": self.insert_ID,
                        "element_idx": self.num_elements + elem_idx,
                        "action_idx": act_idx,
                        "run_idx": 0,
                    }
                    data_idx_i = action.generate_data_index(
                        data_idx=data_idx,
                        workflow=self.workflow,
                        param_source=param_source,
                    )
                    run_0 = {"data_idx": data_idx_i}
                    action_runs[act_idx] = [run_0]
                    data_idx.update(data_idx_i)

            elements.append(
                {
                    "global_idx": self.workflow.num_elements + elem_idx,
                    "index": self.num_elements + elem_idx,
                    "actions": action_runs,
                    "schema_parameters": list(schema_params),
                    "es_idx": self.num_element_sets - 1,
                    "seq_idx": {k: v[elem_idx] for k, v in element_seq_idx.items()},
                    "loop_idx": {},  # TODO
                }
            )

        self.workflow._store.add_elements(self.index, self.insert_ID, elements)
        self._pending_num_elements += len(elements)

        return list(range(*global_element_idx_range))

    def add_elements(
        self,
        base_element=None,
        inputs=None,
        input_files=None,
        sequences=None,
        resources=None,
        repeats=None,
        input_sources=None,
        nesting_order=None,
        element_sets=None,
        sourceable_elements=None,
        propagate_to=None,
        return_indices=False,
    ):
        with self.workflow.batch_update():
            return self._add_elements(
                base_element=base_element,
                inputs=inputs,
                input_files=input_files,
                sequences=sequences,
                resources=resources,
                repeats=repeats,
                input_sources=input_sources,
                nesting_order=nesting_order,
                element_sets=element_sets,
                sourceable_elements=sourceable_elements,
                propagate_to=propagate_to,
                return_indices=return_indices,
            )

    def _add_elements(
        self,
        base_element=None,
        inputs=None,
        input_files=None,
        sequences=None,
        resources=None,
        repeats=None,
        input_sources=None,
        nesting_order=None,
        element_sets=None,
        sourceable_elements=None,
        propagate_to=None,
        return_indices=False,
    ):
        """Add more elements to this task.

        Parameters
        ----------
        sourceable_elements : list of int, optional
            If specified, a list of global element indices from which inputs
            may be sourced. If not specified, all workflow elements are considered
            sourceable.
        propagate_to : list of ElementPropagation, optional
            If specified as an empty or non-empty list, propagate the new elements
            downstream. If an `ElementPropagation` object is not specified for a given
            task, propagation will be attempted using default behaviour.
        return_indices : bool, optional
            If True, return the list of indices of the newly added elements. False by
            default.

        """

        if base_element is not None:
            if base_element.task is not self:
                raise ValueError("If specified, `base_element` must belong to this task.")
            b_inputs, b_resources = base_element.to_element_set_data()
            inputs = inputs or b_inputs
            resources = resources or b_resources

        element_sets = self.app.ElementSet.ensure_element_sets(
            inputs=inputs,
            input_files=input_files,
            sequences=sequences,
            resources=resources,
            repeats=repeats,
            input_sources=input_sources,
            nesting_order=nesting_order,
            element_sets=element_sets,
            sourceable_elements=sourceable_elements,
        )

        elem_idx = []  # global element indices
        for elem_set_i in element_sets:
            elem_set_i = elem_set_i.prepare_persistent_copy()
            elem_idx += self._add_element_set(elem_set_i)

        if propagate_to is not None:

            # TODO: also accept a dict as func arg:
            propagate_to = {i.task.unique_name: i for i in propagate_to}

            for task in self.downstream_tasks:

                elem_propagate = propagate_to.get(
                    task.unique_name, ElementPropagation(task=task)
                )
                if self.unique_name not in (
                    i.unique_name
                    for i in elem_propagate.element_set.get_task_dependencies(
                        as_objects=True
                    )
                ):
                    # TODO: why can't we just do
                    #  `if self in not elem_propagate.element_set.task_dependencies:`?
                    continue

                # TODO: generate a new ElementSet for this task;
                #       Assume for now we use a single base element set.
                #       Later, allow combining multiple element sets.

                elem_set_i = self.app.ElementSet(
                    inputs=elem_propagate.element_set.inputs,
                    input_files=elem_propagate.element_set.input_files,
                    sequences=elem_propagate.element_set.sequences,
                    resources=elem_propagate.element_set.resources,
                    repeats=elem_propagate.element_set.repeats,
                    nesting_order=elem_propagate.nesting_order,
                    sourceable_elements=elem_idx,
                )
                prop_elem_idx = task._add_elements(
                    element_sets=[elem_set_i],
                    return_indices=True,
                )
                elem_idx.extend(prop_elem_idx)

        if return_indices:
            return elem_idx

    def get_task_dependencies(
        self,
        as_objects: bool = False,
    ) -> List[Union[int, WorkflowTask]]:
        """Get tasks (insert ID or WorkflowTask objects) that this task depends on.

        Dependencies may come from either elements from upstream tasks, or from locally
        defined inputs/sequences/defaults from upstream tasks."""

        deps = []
        for element_set in self.template.element_sets:
            for sources in element_set.input_sources.values():
                for src in sources:
                    if (
                        src.source_type is InputSourceType.TASK
                        and src.task_ref not in deps
                    ):
                        deps.append(src.task_ref)

        deps = sorted(deps)
        if as_objects:
            deps = [self.workflow.tasks.get(insert_ID=i) for i in deps]

        return deps

    def get_dependent_tasks(
        self,
        as_objects: bool = False,
    ) -> List[Union[int, WorkflowTask]]:
        """Get tasks (insert ID or WorkflowTask objects) that depends on this task."""
        deps = []
        for task in self.downstream_tasks:
            for element_set in task.template.element_sets:
                for sources in element_set.input_sources.values():
                    for src in sources:
                        if (
                            src.source_type is InputSourceType.TASK
                            and src.task_ref == self.insert_ID
                            and task.insert_ID not in deps
                        ):
                            deps.append(task.insert_ID)
        deps = sorted(deps)
        if as_objects:
            deps = [self.workflow.tasks.get(insert_ID=i) for i in deps]
        return deps

    def get_element_dependencies(
        self,
        as_objects: bool = False,
    ) -> List[Union[Tuple[int, int], Element]]:
        """Get elements from upstream tasks (tuples of (task_insert_ID, element idx) or
        Element objects) that this task depends on."""

        deps = []
        for element in self.elements:
            for (ti_ID, e_idx) in element.get_element_dependencies(as_objects=False):
                if (ti_ID, e_idx) not in deps:
                    deps.append((ti_ID, e_idx))

        deps = sorted(deps)
        if as_objects:
            deps = self.workflow.get_elements_from_indices(deps)

        return deps

    def get_dependent_elements(
        self,
        as_objects: bool = False,
    ) -> List[Union[Tuple[int, int], Element]]:
        """Get elements from downstream tasks (tuples of (task_insert_ID, element idx) or
        Element objects) that depend on this task."""
        deps = []
        for task in self.downstream_tasks:
            for element in task.elements:
                for dep_i in element.get_task_dependencies(as_objects=False):
                    if dep_i == self.insert_ID:
                        key = (task.insert_ID, element.index)
                        if key not in deps:
                            deps.append(key)

        deps = sorted(deps)
        if as_objects:
            deps = self.workflow.get_elements_from_indices(deps)

        return deps

    @property
    def inputs(self):
        return self.app.TaskInputParameters(self)

    @property
    def outputs(self):
        return self.app.TaskOutputParameters(self)

    def get(self, path, raise_on_missing=False, default=None):
        return self.app.Parameters(
            self,
            path=path,
            return_element_parameters=False,
            raise_on_missing=raise_on_missing,
            default=default,
        )

    def _path_to_parameter(self, path):
        if len(path) != 2 or path[0] == "resources":
            return

        if path[0] == "inputs":
            for i in self.template.schemas:
                for j in i.inputs:
                    if j.parameter.typ == path[1]:
                        return j.parameter

        elif path[0] == "outputs":
            for i in self.template.schemas:
                for j in i.outputs:
                    if j.parameter.typ == path[1]:
                        return j.parameter

    def _get_merged_parameter_data(
        self,
        data_index,
        path=None,
        raise_on_missing=False,
        default: Any = None,
    ):
        """Get element data from the persistent store."""

        path = [] if not path else path.split(".")
        parameter = self._path_to_parameter(path)
        current_value = None
        is_cur_val_set = False
        for path_i, data_idx_i in data_index.items():

            path_i = path_i.split(".")
            is_parent = False
            is_update = False
            try:
                rel_path = get_relative_path(path, path_i)
                is_parent = True
            except ValueError:
                try:
                    update_path = get_relative_path(path_i, path)
                    is_update = True

                except ValueError:
                    # no intersection between paths
                    continue

            # TODO: test unset outputs are returned correctly
            data = self.workflow._get_parameter_data(data_idx_i)

            if is_parent:
                # replace current value:
                try:
                    current_value = get_in_container(data, rel_path, cast_indices=True)
                    is_cur_val_set = True
                except (KeyError, IndexError, ValueError):
                    continue

            elif is_update:
                # update sub-part of current value
                current_value = current_value or {}
                set_in_container(current_value, update_path, data, ensure_path=True)
                is_cur_val_set = True

        if not is_cur_val_set:
            if raise_on_missing:
                raise ValueError(f"Path {path} does not exist in the element data.")
            else:
                current_value = default

        if parameter and parameter._value_class:
            current_value = parameter._value_class(**current_value)

        return current_value

    def test_action_rule(self, rule, data_idx):
        check = rule.check_exists or rule.check_missing
        if check:
            param_s = check.split(".")
            if len(param_s) > 2:
                # sub-parameter, so need to retrieve parameter data
                try:
                    self._get_merged_parameter_data(data_idx, raise_on_missing=True)
                    return True if rule.check_exists else False
                except ValueError:
                    return False if rule.check_exists else True
            else:
                if rule.check_exists:
                    return rule.check_exists in data_idx
                elif rule.check_missing:
                    return rule.check_missing not in data_idx

        else:
            # retrieve parameter data
            param_path = ".".join(
                i.condition.callable.kwargs["value"] for i in rule.path.parts[:2]
            )
            element_dat = rule.get(param_path)

            data_path = DataPath(*rule.path.parts[2:])
            rule = Rule(path=data_path, condition=rule.condition, cast=rule.cast)

            return rule.test(element_dat).is_valid


class Elements:

    __slots__ = ("_task",)

    def __init__(self, task: WorkflowTask):

        self._task = task

        # TODO: cache Element objects

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(task={self.task.unique_name!r}, "
            f"num_elements={self.task.num_elements})"
        )

    @property
    def task(self):
        return self._task

    def _get_selection(self, selection):

        if isinstance(selection, int):
            start, stop, step = selection, selection + 1, 1

        elif isinstance(selection, slice):
            start, stop, step = selection.start, selection.stop, selection.step
            stop = self.task.num_elements if stop is None else stop
            start = start or 0
            step = 1 if step is None else step

        selection = slice(start, stop, step)
        length = len(range(*selection.indices(self.task.num_elements)))

        return selection, length

    def islice(
        self,
        start: int = None,
        end: int = None,
    ) -> Iterator[Element]:

        selection, _ = self._get_selection(slice(start, end))
        for i in self.task.workflow.get_task_elements_islice(self.task, selection):
            yield i

    def __len__(self):
        return self.task.num_elements

    def __iter__(self):
        return self.islice()

    def __getitem__(
        self,
        selection: Union[int, slice],
    ) -> Union[Element, List[Element]]:

        selection, length = self._get_selection(selection)
        elements = self.task.workflow.get_task_elements(self.task, selection)

        if length == 1:
            return elements[0]
        else:
            return elements


@dataclass
class Parameters:

    _app_attr = "_app"

    task: WorkflowTask
    path: str
    return_element_parameters: bool
    raise_on_missing: Optional[bool] = False
    default: Optional[Any] = None

    def islice(self, start=None, end=None):
        for i in self.task.workflow.pIO.task_parameter_islice(
            self.task, self.path, start, end
        ):
            yield i

    def _get_selection(self, selection):

        if isinstance(selection, int):
            start, stop, step = selection, selection + 1, 1

        elif isinstance(selection, slice):
            start, stop, step = selection.start, selection.stop, selection.step
            stop = self.task.num_elements if stop is None else stop
            start = start or 0
            step = 1 if step is None else step

        selection = slice(start, stop, step)
        length = len(range(*selection.indices(self.task.num_elements)))

        return selection, length

    def __iter__(self):
        return self.islice()

    def __getitem__(self, selection: Union[int, slice]) -> Union[Any, List[Any]]:

        selection, length = self._get_selection(selection)
        elements = self.task.workflow.get_task_elements(self.task, selection)
        if self.return_element_parameters:
            params = [
                self._app.ElementParameter(
                    task=self.task,
                    path=self.path,
                    parent=self,
                    element=i,
                    data_idx=i.get_data_idx(self.path),
                )
                for i in elements
            ]
        else:
            params = [
                i.get(
                    path=self.path,
                    raise_on_missing=self.raise_on_missing,
                    default=self.default,
                )
                for i in elements
            ]

        if length == 1:
            return params[0]
        else:
            return params


@dataclass
class TaskInputParameters:
    """For retrieving schema input parameters across all elements."""

    _app_attr = "_app"

    task: WorkflowTask

    def __getattr__(self, name):
        if name not in self._get_input_names():
            raise ValueError(f"No input named {name!r}.")
        return self._app.Parameters(self.task, f"inputs.{name}", True)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{', '.join(f'{i!r}' for i in self._get_input_names())})"
        )

    def __dir__(self):
        return super().__dir__() + self._get_input_names()

    def _get_input_names(self):
        return sorted(self.task.template.all_schema_input_types)


@dataclass
class TaskOutputParameters:
    """For retrieving schema output parameters across all elements."""

    _app_attr = "_app"

    task: WorkflowTask

    def __getattr__(self, name):
        if name not in self._get_output_names():
            raise ValueError(f"No output named {name!r}.")
        return self._app.Parameters(self.task, f"outputs.{name}", True)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{', '.join(f'{i!r}' for i in self._get_output_names())})"
        )

    def __dir__(self):
        return super().__dir__() + self._get_output_names()

    def _get_output_names(self):
        return sorted(self.task.template.all_schema_output_types)


@dataclass
class ElementPropagation:
    """Class to represent how a newly added element set should propagate to a given
    downstream task."""

    task: Task
    element_sets: Optional[Union[List[int], List[ElementSet]]] = None
    nesting_order: Optional[Dict] = None

    @property
    def element_set(self):
        # TEMP property; for now just use the first element set as the base:
        return self.task.template.element_sets[0]



from __future__ import annotations
from contextlib import contextmanager
import copy
from pathlib import Path

import shutil
import time
from typing import Any, Dict, Generator, Iterator, List, Optional, Tuple, Union
import numpy as np
import zarr
from numcodecs import MsgPack

from hpcflow.sdk.core.errors import WorkflowNotFoundError
from hpcflow.sdk.core.utils import (
    bisect_slice,
    ensure_in,
    get_in_container,
    get_md5_hash,
    get_relative_path,
    set_in_container,
)
from hpcflow.sdk.persistence import PersistentStore, dropbox_permission_err_retry
from hpcflow.sdk.typing import PathLike


def _encode_numpy_array(obj, type_lookup, path, root_group, arr_path):
    # Might need to generate new group:
    param_arr_group = root_group.require_group(arr_path)
    names = [int(i) for i in param_arr_group.keys()]
    if not names:
        new_idx = 0
    else:
        new_idx = max(names) + 1
    param_arr_group.create_dataset(name=f"arr_{new_idx}", data=obj)
    type_lookup["arrays"].append([path, new_idx])

    return None


def _decode_numpy_arrays(obj, type_lookup, path, arr_group, dataset_copy):
    for arr_path, arr_idx in type_lookup["arrays"]:
        try:
            rel_path = get_relative_path(arr_path, path)
        except ValueError:
            continue

        dataset = arr_group.get(f"arr_{arr_idx}")
        if dataset_copy:
            dataset = dataset[:]

        if rel_path:
            set_in_container(obj, rel_path, dataset)
        else:
            obj = dataset

    return obj


class ZarrPersistentStore(PersistentStore):
    """An efficient storage backend using Zarr that supports parameter-setting from
    multiple processes."""

    _param_grp_name = "parameter_data"
    _elem_grp_name = "element_data"
    _param_base_arr_name = "base"
    _param_sources_arr_name = "sources"
    _param_user_arr_grp_name = "arrays"
    _param_data_arr_grp_name = lambda _, param_idx: f"param_{param_idx}"
    _task_grp_name = lambda _, insert_ID: f"task_{insert_ID}"
    _task_elem_arr_name = "elements"

    _parameter_encoders = {np.ndarray: _encode_numpy_array}  # keys are types
    _parameter_decoders = {"arrays": _decode_numpy_arrays}  # keys are keys in type_lookup

    def __init__(self, workflow: Workflow) -> None:
        self._metadata = None  # cache used in `cached_load` context manager
        super().__init__(workflow)

    def exists(self) -> bool:
        try:
            self._get_root_group()
        except zarr.errors.PathNotFoundError:
            return False
        return True

    @property
    def has_pending(self) -> bool:
        """Returns True if there are pending changes that are not yet committed."""
        return any(bool(v) for k, v in self._pending.items() if k != "element_attrs")

    def _get_pending_dct(self) -> Dict:
        dct = super()._get_pending_dct()
        dct["element_attrs"] = {}  # keys are task indices
        dct["parameter_data"] = 0  # keep number of pending data rather than indices
        return dct

    @classmethod
    def write_empty_workflow(
        cls,
        template_js: Dict,
        template_components_js: Dict,
        path: Path,
        overwrite: bool,
    ) -> None:

        replaced_file = None
        if path.exists():
            if overwrite:
                replaced_file = cls._rename_existing(path)
            else:
                raise ValueError(f"Path already exists: {path}.")

        metadata = {
            "template": template_js,
            "template_components": template_components_js,
            "num_added_tasks": 0,
        }
        if replaced_file:
            metadata["replaced_file"] = str(replaced_file.name)

        store = zarr.DirectoryStore(path)
        root = zarr.group(store=store, overwrite=overwrite)
        root.attrs.update(metadata)

        root.create_group(name=cls._elem_grp_name)
        parameter_data = root.create_group(name=cls._param_grp_name)
        parameter_data.create_dataset(
            name=cls._param_base_arr_name,
            shape=0,
            dtype=object,
            object_codec=MsgPack(),
            chunks=1,
        )
        parameter_data.create_dataset(
            name=cls._param_sources_arr_name,
            shape=0,
            dtype=object,
            object_codec=MsgPack(),
            chunks=1000,  # TODO: check this is a sensible size with many parameters
        )
        parameter_data.create_group(name=cls._param_user_arr_grp_name)

    def load_metadata(self):
        return self._metadata or self._load_metadata()

    def _load_metadata(self):
        return self._get_root_group(mode="r").attrs.asdict()

    @contextmanager
    def cached_load(self) -> Iterator[Dict]:
        """Context manager to cache the root attributes (i.e. metadata)."""
        if self._metadata:
            yield
        else:
            try:
                self._metadata = self._load_metadata()
                yield
            finally:
                self._metadata = None

    def _get_root_group(self, mode: str = "r") -> zarr.Group:
        return zarr.open(self.workflow.path, mode=mode)

    def _get_parameter_group(self, mode: str = "r") -> zarr.Group:
        return self._get_root_group(mode=mode).get(self._param_grp_name)

    def _get_parameter_base_array(self, mode: str = "r") -> zarr.Array:
        return self._get_parameter_group(mode=mode).get(self._param_base_arr_name)

    def _get_parameter_sources_array(self, mode: str = "r") -> zarr.Array:
        return self._get_parameter_group(mode=mode).get(self._param_sources_arr_name)

    def _get_parameter_user_array_group(self, mode: str = "r") -> zarr.Group:
        return self._get_parameter_group(mode=mode).get(self._param_user_arr_grp_name)

    def _get_parameter_data_array_group(
        self,
        parameter_idx: int,
        mode: str = "r",
    ) -> zarr.Group:
        return self._get_parameter_user_array_group(mode=mode).get(
            self._param_data_arr_grp_name(parameter_idx)
        )

    def _get_element_group(self, mode: str = "r") -> zarr.Group:
        return self._get_root_group(mode=mode).get(self._elem_grp_name)

    def _get_task_group_path(self, insert_ID: int) -> str:
        return self._task_grp_name(insert_ID)

    def _get_task_group(self, insert_ID: int, mode: str = "r") -> zarr.Group:
        return self._get_element_group(mode=mode).get(self._task_grp_name(insert_ID))

    def _get_task_elements_array(self, insert_ID: int, mode: str = "r") -> zarr.Array:
        return self._get_task_group(insert_ID, mode=mode).get(self._task_elem_arr_name)

    def _get_task_element_attrs(self, task_idx: int, task_insert_ID: int) -> Dict:
        if task_idx in self._pending["element_attrs"]:
            attrs = self._pending["element_attrs"][task_idx]
        elif task_idx in self._pending["tasks"]:
            # the task is new and not yet committed
            attrs = self._get_element_array_empty_attrs()
        else:
            attrs = self._get_task_elements_array(task_insert_ID, mode="r").attrs.asdict()
        return attrs

    def add_elements(
        self,
        task_idx: int,
        task_insert_ID: int,
        elements: List[Dict],
    ) -> None:

        attrs_original = self._get_task_element_attrs(task_idx, task_insert_ID)
        elements, attrs = self._compress_elements(elements, attrs_original)
        if attrs != attrs_original:
            if task_idx not in self._pending["element_attrs"]:
                self._pending["element_attrs"][task_idx] = {}
            self._pending["element_attrs"][task_idx].update(attrs)

        return super().add_elements(task_idx, task_insert_ID, elements)

    def _compress_elements(self, elements: List, attrs: Dict) -> Tuple[List, Dict]:
        """The great listification of election action run data.

        See also: `_decompress_elements` for the inverse operation.

        """

        attrs = copy.deepcopy(attrs)
        compressed = []
        for elem in elements:
            seq_idx = [
                [ensure_in(k, attrs["sequences"]), v] for k, v in elem["seq_idx"].items()
            ]
            loop_idx = [
                [ensure_in(k, attrs["loops"]), v] for k, v in elem["loop_idx"].items()
            ]
            schema_params = [
                ensure_in(k, attrs["schema_parameters"])
                for k in elem["schema_parameters"]
            ]
            act_runs = []
            for act_idx, runs in elem["actions"].items():
                act_run_i = [
                    act_idx,
                    [
                        [
                            [ensure_in(dk, attrs["parameter_paths"]), dv]
                            for dk, dv in r["data_idx"].items()
                        ]
                        for r in runs
                    ],
                ]
                act_runs.append(act_run_i)

            compact = [
                elem["global_idx"],
                elem["index"],
                elem["es_idx"],
                seq_idx,
                loop_idx,
                schema_params,
                act_runs,
            ]
            compressed.append(compact)
        return compressed, attrs

    def _decompress_elements(self, elements: List, attrs: Dict) -> List:

        out = []
        for elem in elements:
            actions = {}
            for (act_idx, runs) in elem[6]:
                actions[act_idx] = [
                    {
                        "data_idx": {
                            attrs["parameter_paths"][k]: v for (k, v) in data_idx
                        },
                    }
                    for data_idx in runs
                ]
            elem_i = {
                "global_idx": elem[0],
                "index": elem[1],
                "es_idx": elem[2],
                "seq_idx": {attrs["sequences"][k]: v for (k, v) in elem[3]},
                "loop_idx": {attrs["loops"][k]: v for (k, v) in elem[4]},
                "schema_parameters": [attrs["schema_parameters"][k] for k in elem[5]],
                "actions": actions,
            }
            out.append(elem_i)
        return out

    @staticmethod
    def _get_element_array_empty_attrs() -> Dict:
        return {
            "sequences": [],
            "loops": [],
            "schema_parameters": [],
            "parameter_paths": [],
            "input_source_paths": [],
            "source_types": [],
        }

    def _get_zarr_store(self):
        return self._get_root_group().store

    def _remove_pending_parameter_data(self) -> None:
        """Delete pending parameter data from disk."""
        base = self._get_parameter_base_array(mode="r+")
        sources = self._get_parameter_sources_array(mode="r+")
        for param_idx in range(self._pending["parameter_data"], 0, -1):
            grp = self._get_parameter_data_array_group(param_idx - 1)
            if grp:
                zarr.storage.rmdir(store=self._get_zarr_store(), path=grp.path)
        base.resize(base.size - self._pending["parameter_data"])
        sources.resize(sources.size - self._pending["parameter_data"])

    def reject_pending(self) -> None:
        if self._pending["parameter_data"]:
            self._remove_pending_parameter_data()
        super().reject_pending()

    def commit_pending(self) -> None:

        md = self.load_metadata()

        # merge new tasks:
        for task_idx, task_js in self._pending["template_tasks"].items():
            md["template"]["tasks"].insert(task_idx, task_js)  # TODO should be index?

        # write new workflow tasks to disk:
        for task_idx, _ in self._pending["tasks"].items():

            insert_ID = self._pending["template_tasks"][task_idx]["insert_ID"]
            task_group = self._get_element_group(mode="r+").create_group(
                self._get_task_group_path(insert_ID)
            )
            element_arr = task_group.create_dataset(
                name=self._task_elem_arr_name,
                shape=0,
                dtype=object,
                object_codec=MsgPack(),
                chunks=1000,  # TODO: check this is a sensible size with many elements
            )
            element_arr.attrs.update(self._get_element_array_empty_attrs())
            md["num_added_tasks"] += 1

        # merge new template components:
        self._merge_pending_template_components(md["template_components"])

        # merge new element sets:
        for task_idx, es_js in self._pending["element_sets"].items():
            md["template"]["tasks"][task_idx]["element_sets"].extend(es_js)

        # write new elements to disk:
        for (task_idx, insert_ID), elements in self._pending["elements"].items():
            elem_arr = self._get_task_elements_array(insert_ID, mode="r+")
            elem_arr_add = np.empty((len(elements)), dtype=object)
            elem_arr_add[:] = elements
            elem_arr.append(elem_arr_add)
            if task_idx in self._pending["element_attrs"]:
                elem_arr.attrs.put(self._pending["element_attrs"][task_idx])

        if self._pending["remove_replaced_file_record"]:
            del md["replaced_file"]

        # commit updated metadata:
        self._get_root_group(mode="r+").attrs.put(md)
        self.clear_pending()

    def _get_persistent_template_components(self) -> Dict:
        return self.load_metadata()["template_components"]

    def get_template(self) -> Dict:
        # No need to consider pending; this is called once per Workflow object
        return self.load_metadata()["template"]

    def get_num_added_tasks(self) -> int:
        return self.load_metadata()["num_added_tasks"] + len(self._pending["tasks"])

    def get_all_tasks_metadata(self) -> List[Dict]:
        out = []
        for _, grp in self._get_element_group().groups():
            elem_arr = grp.get(self._task_elem_arr_name)
            out.append({"num_elements": len(elem_arr)})
        return out

    def get_task_elements(
        self,
        task_idx: int,
        task_insert_ID: int,
        selection: slice,
    ) -> List:

        num_pers = self.workflow.tasks[task_idx]._num_elements
        pers_slice, pend_slice = bisect_slice(selection, num_pers)
        pers_range = range(pers_slice.start, pers_slice.stop, pers_slice.step)

        if len(pers_range):
            elem_arr = self._get_task_elements_array(task_insert_ID)
            try:
                elements = list(elem_arr[pers_slice])
            except zarr.errors.NegativeStepError:
                elements = [elem_arr[idx] for idx in pers_range]
        else:
            elements = []

        key = (task_idx, task_insert_ID)
        if key in self._pending["elements"]:
            elements += self._pending["elements"][key][pend_slice]

        return self._decompress_elements(elements, self._get_task_element_attrs(*key))

    def _encode_parameter_data(
        self,
        obj: Any,
        root_group: zarr.Group,
        arr_path: str,
        path: List = None,
        type_lookup: Optional[Dict] = None,
    ) -> Dict[str, Any]:

        return super()._encode_parameter_data(
            obj=obj,
            path=path,
            type_lookup=type_lookup,
            root_group=root_group,
            arr_path=arr_path,
        )

    def _decode_parameter_data(
        self,
        data: Union[None, Dict],
        arr_group: zarr.Group,
        path: Optional[List[str]] = None,
        dataset_copy=False,
    ) -> Any:

        return super()._decode_parameter_data(
            data=data,
            path=path,
            arr_group=arr_group,
            dataset_copy=dataset_copy,
        )

    def _add_parameter_data(self, data: Any, source: Dict) -> int:

        base_arr = self._get_parameter_base_array(mode="r+")
        sources = self._get_parameter_sources_array(mode="r+")
        idx = base_arr.size

        if data is not None:
            data = self._encode_parameter_data(
                obj=data["data"],
                root_group=self._get_parameter_user_array_group(mode="r+"),
                arr_path=self._param_data_arr_grp_name(idx),
            )

        base_arr.append([data])
        sources.append([source])
        self._pending["parameter_data"] += 1
        self.save()

        return idx

    def set_parameter(self, index: int, data: Any) -> None:
        """Set the value of a pre-allocated parameter."""

        if self.is_parameter_set(index):
            raise RuntimeError(f"Parameter at index {index} is already set!")

        base_arr = self._get_parameter_base_array(mode="r+")

        data, type_lookup = self._encode_parameter_data(
            obj=data,
            root_group=self._get_parameter_user_array_group(mode="r+"),
            arr_path=self._param_data_arr_grp_name(index),
        )

        base_arr[index] = {"data": data, "type_lookup": type_lookup}

    def _get_parameter_data(self, index: int) -> Any:
        return self._get_parameter_base_array(mode="r")[index]

    def get_parameter_data(self, index: int) -> Any:
        return self._decode_parameter_data(
            data=self._get_parameter_data(index),
            arr_group=self._get_parameter_data_array_group(index),
        )

    def get_parameter_source(self, index: int) -> Dict:
        return self._get_parameter_sources_array(mode="r")[index]

    def get_all_parameter_data(self) -> Dict[int, Any]:
        max_key = self._get_parameter_base_array(mode="r").size - 1
        out = {}
        for idx in range(max_key + 1):
            out[idx] = self.get_parameter_data(idx)
        return out

    def is_parameter_set(self, index: int) -> bool:
        return self._get_parameter_data(index) is not None

    def check_parameters_exist(
        self, indices: Union[int, List[int]]
    ) -> Union[bool, List[bool]]:
        is_multi = True
        if not isinstance(indices, (list, tuple)):
            is_multi = False
            indices = [indices]
        base = self._get_parameter_base_array(mode="r")
        idx_range = range(base.size)
        exists = [i in idx_range for i in indices]
        if not is_multi:
            exists = exists[0]
        return exists

    @dropbox_permission_err_retry
    def delete_no_confirm(self) -> None:
        """Permanently delete the workflow data with no confirmation."""
        # Dropbox (on Windows, at least) seems to try to re-sync some of the workflow
        # files if it is deleted soon after creation, which is the case on a failed
        # workflow creation (e.g. missing inputs):
        while self.workflow.path.is_dir():
            shutil.rmtree(self.workflow.path)
            time.sleep(0.5)

    @dropbox_permission_err_retry
    def remove_replaced_file(self) -> None:
        md = self.load_metadata()
        if "replaced_file" in md:
            shutil.rmtree(Path(md["replaced_file"]))
            self._pending["remove_replaced_file_record"] = True
            self.save()

    @dropbox_permission_err_retry
    def reinstate_replaced_file(self) -> None:
        print(f"reinstate replaced file!")
        md = self.load_metadata()
        if "replaced_file" in md:
            # TODO does rename work for the dir?
            Path(md["replaced_file"]).rename(self.path)

    def copy(self, path: PathLike = None) -> None:
        shutil.copytree(self.path, path)

    def is_modified_on_disk(self) -> bool:
        if self._metadata:
            return get_md5_hash(self._load_metadata()) != get_md5_hash(self._metadata)
        else:
            # nothing to compare to
            return False



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

from valida.conditions import ConditionLike

from hpcflow.sdk.core.utils import check_valid_py_identifier


class _ElementPrefixedParameter:

    _app_attr = "_app"

    def __init__(
        self,
        prefix: str,
        element: Optional[Element] = None,
        element_action: Optional[ElementAction] = None,
        element_action_run: Optional[ElementActionRun] = None,
    ) -> None:

        self._prefix = prefix
        self._element = element
        self._element_action = element_action
        self._element_action_run = element_action_run

    @property
    def _parent(self):
        return self._element or self._element_action or self._element_action_run

    @property
    def _element_obj(self):
        return (
            self._element
            or self._element_action.element
            or self._element_action_run.element
        )

    @property
    def _task(self):
        return self._parent.task

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{', '.join(f'{i!r}' for i in self._get_prefixed_names())})"
        )

    def _get_prefixed_names(self):
        return sorted(self._parent.get_parameter_names(self._prefix))

    def __getattr__(self, name):
        if name not in self._get_prefixed_names():
            raise ValueError(f"No {self._prefix} named {name!r}.")

        data_idx = self._parent.get_data_idx(path=f"{self._prefix}.{name}")
        param = self._app.ElementParameter(
            path=f"{self._prefix}.{name}",
            task=self._task,
            data_idx=data_idx,
            parent=self._parent,
            element=self._element_obj,
        )
        return param

    def __dir__(self):
        return super().__dir__() + self._get_prefixed_names()


class ElementInputs(_ElementPrefixedParameter):
    def __init__(
        self,
        element: Optional[Element] = None,
        element_action: Optional[ElementAction] = None,
        element_action_run: Optional[ElementActionRun] = None,
    ) -> None:
        super().__init__("inputs", element, element_action, element_action_run)


class ElementOutputs(_ElementPrefixedParameter):
    def __init__(
        self,
        element: Optional[Element] = None,
        element_action: Optional[ElementAction] = None,
        element_action_run: Optional[ElementActionRun] = None,
    ) -> None:
        super().__init__("outputs", element, element_action, element_action_run)


class ElementInputFiles(_ElementPrefixedParameter):
    def __init__(
        self,
        element: Optional[Element] = None,
        element_action: Optional[ElementAction] = None,
        element_action_run: Optional[ElementActionRun] = None,
    ) -> None:
        super().__init__("input_files", element, element_action, element_action_run)


class ElementOutputFiles(_ElementPrefixedParameter):
    def __init__(
        self,
        element: Optional[Element] = None,
        element_action: Optional[ElementAction] = None,
        element_action_run: Optional[ElementActionRun] = None,
    ) -> None:
        super().__init__("output_files", element, element_action, element_action_run)


@dataclass
class ElementResources:
    scratch: str
    num_cores: int

    def __post_init__(self):
        if self.num_cores is None:
            self.num_cores = 1


class Element:

    _app_attr = "app"

    # TODO: use slots

    def __init__(
        self,
        task: WorkflowTask,
        index: int,
        actions: List[Dict],
        global_idx: int,
        es_idx: int,
        seq_idx: Dict,
        schema_parameters: List[str],
        loop_idx: Optional[Dict],
    ) -> None:

        self._task = task
        self._global_idx = global_idx
        self._index = index
        self._es_idx = es_idx
        self._seq_idx = seq_idx
        self._loop_idx = loop_idx
        self._schema_parameters = schema_parameters
        self._actions = actions

        # assigned on first access of corresponding properties:
        self._inputs = None
        self._outputs = None
        self._input_files = None
        self._output_files = None
        self._action_objs = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"task={self.task.unique_name!r}, index={self.index!r}"
            f")"
        )

    @property
    def task(self) -> WorkflowTask:
        return self._task

    @property
    def global_idx(self) -> int:
        return self._global_idx

    @property
    def index(self) -> int:
        """Get the index of the element within the task.

        Note: the `global_idx` attribute returns the index of the element within the
        workflow, across all tasks."""

        return self._index

    @property
    def element_set_idx(self) -> int:
        return self._es_idx

    @property
    def element_set(self):
        return self.task.template.element_sets[self.element_set_idx]

    @property
    def sequence_idx(self) -> Dict[str, int]:
        return self._seq_idx

    @property
    def loop_idx(self) -> Dict[str, int]:
        return self._loop_idx

    @property
    def schema_parameters(self) -> List[str]:
        return self._schema_parameters

    @property
    def actions(self) -> Dict[ElementAction]:
        if self._action_objs is None:
            self._action_objs = {
                act_idx: self.app.ElementAction(
                    element=self,
                    action_idx=act_idx,
                    runs=runs,
                )
                for act_idx, runs in self._actions.items()
            }
        return self._action_objs

    @property
    def action_runs(self) -> List[ElementActionRun]:
        """Get a list of element action runs, where only the final run is taken for each
        element action."""
        return [i.runs[-1] for i in self.actions.values()]

    @property
    def workflow(self) -> Workflow:
        return self.task.workflow

    @property
    def inputs(self) -> ElementInputs:
        if not self._inputs:
            self._inputs = self.app.ElementInputs(element=self)
        return self._inputs

    @property
    def outputs(self) -> ElementOutputs:
        if not self._outputs:
            self._outputs = self.app.ElementOutputs(element=self)
        return self._outputs

    @property
    def input_files(self) -> ElementInputFiles:
        if not self._input_files:
            self._input_files = self.app.ElementInputFiles(element=self)
        return self._input_files

    @property
    def output_files(self) -> ElementOutputFiles:
        if not self._output_files:
            self._output_files = self.app.ElementOutputFiles(element=self)
        return self._output_files

    @property
    def dir_name(self):
        return str(self.index)

    @property
    def dir_path(self):
        return self.task.dir_path / self.dir_name

    def to_element_set_data(self):
        """Generate lists of workflow-bound InputValues and ResourceList."""
        inputs = []
        resources = []
        for k, v in self.get_data_idx().items():

            k_s = k.split(".")

            if k_s[0] == "inputs":
                inp_val = self.app.InputValue(
                    parameter=k_s[1],
                    path=k_s[2:] or None,
                    value=None,
                )
                inp_val._value_group_idx = v
                inp_val._workflow = self.workflow
                inputs.append(inp_val)

            elif k_s[0] == "resources":
                scope = self.app.ActionScope.from_json_like(k_s[1])
                res = self.app.ResourceSpec(scope=scope)
                res._value_group_idx = v
                res._workflow = self.workflow
                resources.append(res)

        return inputs, resources

    def get_sequence_value(self, sequence_path: str) -> Any:
        seq = self.element_set.get_sequence_from_path(sequence_path)
        if not seq:
            raise ValueError(
                f"No sequence with path {sequence_path!r} in this element's originating "
                f"element set."
            )
        return seq.values[self.sequence_idx[sequence_path]]

    def get_parameter_names(self, prefix: str) -> List[str]:
        return list(
            ".".join(i.split(".")[1:])
            for i in self.schema_parameters
            if i.startswith(prefix)
        )

    # def _path_to_parameter(self, path):
    #     TODO: remove
    #     if len(path) != 2 or path[0] == "resources":
    #         return

    #     if path[0] == "inputs":
    #         for i in self.task.template.schemas:
    #             for j in i.inputs:
    #                 if j.parameter.typ == path[1]:
    #                     return j.parameter

    #     elif path[0] == "outputs":
    #         for i in self.task.template.schemas:
    #             for j in i.outputs:
    #                 if j.parameter.typ == path[1]:
    #                     return j.parameter

    def get_data_idx(
        self,
        path: str = None,
        action_idx: int = None,
        run_idx: int = -1,
    ) -> Dict[str, int]:
        """
        Parameters
        ----------
        action_idx
            The index of the action within the schema.
        """

        if action_idx is None:
            # default behaviour if no action_idx is specified: inputs should be from first
            # action where that input is defined; outputs should include modifications
            # from all actions. TODO: what about resources?
            data_idx = {}
            for action in self.actions.values():
                data_idx.update(action.runs[run_idx].data_idx)
                if path and "inputs" in path and path in data_idx:
                    break
        else:
            elem_act = self.actions[action_idx]
            data_idx = elem_act.runs[run_idx].data_idx

        if path:
            data_idx = {k: v for k, v in data_idx.items() if k.startswith(path)}

        return data_idx

    def get_parameter_sources(
        self,
        path: str = None,
        action_idx: int = None,
        run_idx: int = -1,
        typ: str = None,
        as_strings: bool = False,
        use_task_index: bool = False,
    ) -> Dict[str, Union[str, Dict[str, Any]]]:
        """
        Parameters
        ----------
        use_task_index
            If True, use the task index within the workflow, rather than the task insert
            ID.
        """
        data_idx = self.get_data_idx(path, action_idx, run_idx)
        out = {k: self.workflow._get_parameter_source(v) for k, v in data_idx.items()}
        task_key = "task_insert_ID"

        if use_task_index:
            task_key = "task_idx"
            out_task_idx = {}
            for k, v in out.items():
                insert_ID = v.pop("task_insert_ID", None)
                if insert_ID is not None:
                    v[task_key] = self.workflow.tasks.get(insert_ID=insert_ID).index
                out_task_idx[k] = v
            out = out_task_idx

        if typ:
            out = {k: v for k, v in out.items() if v["type"] == typ}

        if as_strings:
            # format as a dict with compact string values
            self_task_val = (
                self.task.index if task_key == "task_idx" else self.task.insert_ID
            )
            out_strs = {}
            for k, v in out.items():
                if v["type"] == "local_input":
                    if v[task_key] == self_task_val:
                        out_strs[k] = "local"
                    else:
                        out_strs[k] = f"task.{v[task_key]}.input"
                elif v["type"] == "default_input":
                    out_strs == "default"
                else:
                    out_strs[k] = (
                        f"task.{v[task_key]}.element.{v['element_idx']}."
                        f"action.{v['action_idx']}.run.{v['run_idx']}"
                    )
            out = out_strs

        return out

    def get(
        self,
        path: str = None,
        action_idx: int = None,
        run_idx: int = -1,
        default: Any = None,
        raise_on_missing: bool = False,
    ) -> Any:
        """Get element data from the persistent store."""
        # TODO include a "stats" parameter which when set we know the run has been
        # executed (or if start time is set but not end time, we know it's running or
        # failed.)
        return self.task._get_merged_parameter_data(
            data_index=self.get_data_idx(action_idx=action_idx, run_idx=run_idx),
            path=path,
            raise_on_missing=raise_on_missing,
            default=default,
        )

    def get_EAR_dependencies(
        self,
        as_objects: Optional[bool] = False,
    ) -> List[Union[Tuple[int, int, int, int], ElementActionRun]]:
        """Get EARs that this element depends on (excluding EARs of this element)."""
        out = sorted(
            set(
                j
                for i in self.action_runs
                for j in i.get_EAR_dependencies(as_objects=False)
                if not (j[0] == self.task.insert_ID and j[1] == self.index)
            )
        )
        if as_objects:
            out = self.workflow.get_EARs_from_indices(out)
        return out

    def get_element_dependencies(
        self,
        as_objects: Optional[bool] = False,
    ) -> List[Union[Tuple[int, int], Element]]:
        """Get elements that this element depends on."""
        # TODO: this will be used in viz.
        out = sorted(
            set((i[0], i[1]) for i in self.get_EAR_dependencies(as_objects=False))
        )
        if as_objects:
            out = self.workflow.get_elements_from_indices(out)
        return out

    def get_input_dependencies(self) -> Dict[str, Dict]:
        """Get locally defined inputs/sequences/defaults from other tasks that this
        element depends on."""
        out = {}
        for k, v in self.get_parameter_sources().items():
            if (
                v["type"] in ["local_input", "default_input"]
                and v["task_insert_ID"] != self.task.insert_ID
            ):
                out[k] = v

        return out

    def get_task_dependencies(
        self, as_objects: bool = False
    ) -> List[Union[int, WorkflowTask]]:
        """Get tasks (insert ID or WorkflowTask objects) that this element depends on.

        Dependencies may come from either elements from upstream tasks, or from locally
        defined inputs/sequences/defaults from upstream tasks."""

        out = []
        for elem_dep in self.get_element_dependencies(as_objects=False):
            out.append(elem_dep[0])

        for i in self.get_input_dependencies().values():
            out.append(i["task_insert_ID"])

        out = sorted(set(out))

        if as_objects:
            out = [self.workflow.tasks.get(insert_ID=i) for i in out]

        return out

    def get_dependent_EARs(
        self, as_objects: bool = False
    ) -> List[Union[Tuple[int, int, int, int], ElementActionRun]]:
        """Get EARs of downstream tasks that depend on this element."""
        deps = []
        for task in self.task.downstream_tasks:
            for element in task.elements:
                for EAR_i in element.action_runs:
                    for dep_i in EAR_i.get_EAR_dependencies(as_objects=False):
                        if (
                            dep_i[0] == self.task.insert_ID
                            and dep_i[1] == self.index
                            and dep_i not in deps
                        ):
                            deps.append(dep_i)

        deps = sorted(deps)
        if as_objects:
            deps = self.workflow.get_EARs_from_indices(deps)

        return deps

    def get_dependent_elements(
        self,
        as_objects: bool = False,
    ) -> List[Union[Tuple[int, int], Element]]:
        """Get elements of downstream tasks that depend on this element."""
        deps = []
        for task in self.task.downstream_tasks:
            for element in task.elements:
                for dep_i in element.get_element_dependencies(as_objects=False):
                    (ti_ID, e_idx) = dep_i
                    dep_elem = (task.insert_ID, element.index)
                    if (
                        ti_ID == self.task.insert_ID
                        and e_idx == self.index
                        and dep_elem not in deps
                    ):
                        deps.append(dep_elem)

        deps = sorted(deps)
        if as_objects:
            deps = self.workflow.get_elements_from_indices(deps)

        return deps

    def get_dependent_tasks(
        self,
        as_objects: bool = False,
    ) -> List[Union[int, WorkflowTask]]:
        """Get downstream tasks that depend on this element."""
        deps = []
        for task in self.task.downstream_tasks:
            for dep_i in task.get_element_dependencies(as_objects=False):
                (ti_ID, e_idx) = dep_i
                if (
                    ti_ID == self.task.insert_ID
                    and e_idx == self.index
                    and task.insert_ID not in deps
                ):
                    deps.append(task.insert_ID)

        deps = sorted(deps)
        if as_objects:
            deps = [self.workflow.tasks.get(insert_ID=i) for i in deps]

        return deps


@dataclass
class ElementParameter:

    # TODO: do we need `parent` attribute?

    _app_attr = "app"

    task: WorkflowTask
    path: str
    parent: Union[Element, ElementAction, ElementActionRun, Parameters]
    element: Element
    data_idx: Dict[str, int]

    @property
    def value(self):
        return self.task._get_merged_parameter_data(self.data_idx, self.path)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(element={self.element!r}, path={self.path!r})"

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, self.__class__):
            return False
        if self.task == __o.task and self.path == __o.path:
            return True

    @property
    def data_idx_is_set(self):
        return {
            k: self.task.workflow.is_parameter_set(v) for k, v in self.data_idx.items()
        }

    @property
    def is_set(self):
        return all(self.data_idx_is_set.values())

    def get_size(self, **store_kwargs):
        raise NotImplementedError


@dataclass
class ElementFilter:

    parameter_path: ParameterPath
    condition: ConditionLike


@dataclass
class ElementGroup:

    name: str
    where: Optional[ElementFilter] = None
    group_by_distinct: Optional[ParameterPath] = None

    def __post_init__(self):
        self.name = check_valid_py_identifier(self.name)


@dataclass
class ElementRepeats:

    number: int
    where: Optional[ElementFilter] = None



from __future__ import annotations

import copy
import json
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
import shutil
from typing import Any, Dict, Generator, Iterator, List, Optional, Union
from hpcflow.sdk.core.errors import WorkflowNotFoundError
from hpcflow.sdk.core.utils import get_md5_hash

from hpcflow.sdk.persistence import PersistentStore, dropbox_permission_err_retry


class JSONPersistentStore(PersistentStore):
    """A verbose but inefficient storage backend, to help with understanding and
    debugging."""

    def __init__(self, workflow: Workflow) -> None:
        self._loaded = None  # cache used in `cached_load` context manager
        super().__init__(workflow)

    def exists(self) -> bool:
        return self.path.is_file()

    @classmethod
    def write_empty_workflow(
        cls,
        template_js: Dict,
        template_components_js: Dict,
        path: Path,
        overwrite: bool,
    ) -> None:

        replaced_file = None
        if path.exists():
            if overwrite:
                replaced_file = cls._rename_existing(path)
            else:
                raise ValueError(f"Path already exists: {path}.")
        dat = {
            "parameter_data": {},
            "parameter_sources": {},
            "template_components": template_components_js,
            "template": template_js,
            "tasks": [],
            "num_added_tasks": 0,
        }
        if replaced_file:
            dat["replaced_file"] = str(replaced_file.name)

        cls._dump_to_path(path, dat)

    @contextmanager
    def cached_load(self) -> Iterator[Dict]:
        """Context manager to cache the whole JSON document, allowing for multiple read
        operations with one disk read."""
        if self._loaded:
            yield
        else:
            try:
                self._loaded = self._load()
                yield
            finally:
                self._loaded = None

    def _load(self) -> Dict:
        with open(self.path, "rt") as fp:
            return json.load(fp)

    def load(self) -> Dict:
        return self._loaded or self._load()

    @staticmethod
    @dropbox_permission_err_retry
    def _dump_to_path(path, data: Dict) -> None:
        with open(path, "wt", newline="") as fp:
            json.dump(data, fp, indent=4)

    def _dump(self, wk_data: Dict) -> None:
        self._dump_to_path(self.path, wk_data)

    def _add_parameter_data(self, data: Any, source: Dict) -> int:

        idx = len(self.load()["parameter_data"]) + len(self._pending["parameter_data"])

        if data is not None:
            data = self._encode_parameter_data(data["data"])

        self._pending["parameter_data"][idx] = data
        self._pending["parameter_sources"][idx] = source
        self.save()

        return idx

    def set_parameter(self, index: int, data: Any) -> None:
        """Set the value of a pre-allocated parameter."""
        if self.is_parameter_set(index):
            raise RuntimeError(f"Parameter at index {index} is already set!")
        self._pending["parameter_data"][index] = self._encode_parameter_data(data)
        self.save()

    def get_parameter_data(self, index: int) -> Any:
        if index in self._pending["parameter_data"]:
            data = self._pending["parameter_data"][index]
        else:
            data = self.load()["parameter_data"][str(index)]
        return self._decode_parameter_data(data=data)

    def get_parameter_source(self, index: int) -> Dict:
        # TODO: check pending
        return self.load()["parameter_sources"][str(index)]

    def get_all_parameter_data(self) -> Dict[int, Any]:
        if self._pending["parameter_data"]:
            max_key = max(self._pending["parameter_data"].keys())
        else:
            max_key = int(max(self.load()["parameter_data"].keys(), key=lambda x: int(x)))

        out = {}
        for idx in range(max_key + 1):
            out[idx] = self.get_parameter_data(idx)

        return out

    def is_parameter_set(self, index: int) -> bool:
        return self.load()["parameter_data"][str(index)] is not None

    def check_parameters_exist(
        self, indices: Union[int, List[int]]
    ) -> Union[bool, List[bool]]:
        is_multi = True
        if not isinstance(indices, (list, tuple)):
            is_multi = False
            indices = [indices]
        exists = [
            i in self._pending["parameter_data"]
            or str(i) in self.load()["parameter_data"]
            for i in indices
        ]
        if not is_multi:
            exists = exists[0]
        return exists

    def commit_pending(self) -> None:

        wk_data = self.load()

        # commit new tasks:
        for new_index, task_js in self._pending["template_tasks"].items():
            wk_data["template"]["tasks"].insert(new_index, task_js)

        # commit new workflow tasks:
        for new_index, wk_task in self._pending["tasks"].items():
            wk_data["tasks"].insert(new_index, wk_task)
            wk_data["num_added_tasks"] += 1

        # commit new template components:
        self._merge_pending_template_components(wk_data["template_components"])

        # commit new element sets:
        for task_idx, es_js in self._pending["element_sets"].items():
            wk_data["template"]["tasks"][task_idx]["element_sets"].extend(es_js)

        # commit new elements:
        for (task_idx, _), elements in self._pending["elements"].items():
            wk_data["tasks"][task_idx]["elements"].extend(elements)

        # commit new parameters:
        for param_idx, param_dat in self._pending["parameter_data"].items():
            wk_data["parameter_data"][param_idx] = param_dat

        for param_idx, param_src in self._pending["parameter_sources"].items():
            wk_data["parameter_sources"][param_idx] = param_src

        if self._pending["remove_replaced_file_record"]:
            del wk_data["replaced_file"]

        self._dump(wk_data)
        self.clear_pending()

    def _get_persistent_template_components(self) -> Dict:
        return self.load()["template_components"]

    def get_template(self) -> Dict:
        # No need to consider pending; this is called once per Workflow object
        return self.load()["template"]

    def get_num_added_tasks(self) -> int:
        return self.load()["num_added_tasks"] + len(self._pending["tasks"])

    def get_all_tasks_metadata(self) -> List[Dict]:
        # No need to consider pending; this is called once per Workflow object
        return [{"num_elements": len(i["elements"])} for i in self.load()["tasks"]]

    def get_task_elements(
        self,
        task_idx: int,
        task_insert_ID: int,
        selection: slice,
    ) -> List[Dict]:

        if task_idx in self._pending["tasks"]:
            wk_task_dat = self._pending["tasks"][task_idx]
        else:
            wk_task_dat = self.load()["tasks"][task_idx]

        wk_task = copy.deepcopy(wk_task_dat)
        key = (task_idx, task_insert_ID)
        if key in self._pending["elements"]:
            wk_task["elements"] += copy.deepcopy(self._pending["elements"][key])

        # TODO: maybe support slicing beyond end of available elements, like a list?
        #       using slice.indices ?
        elements = []
        for idx in range(selection.start, selection.stop, selection.step):
            for act_idx_str in list(wk_task["elements"][idx]["actions"].keys()):
                runs = wk_task["elements"][idx]["actions"].pop(act_idx_str)
                wk_task["elements"][idx]["actions"][int(act_idx_str)] = runs
            elements.append(wk_task["elements"][idx])

        return elements

    def get_task_parameters(self, task_idx: int, selection: slice) -> List[Dict]:
        pass

    @dropbox_permission_err_retry
    def delete_no_confirm(self) -> None:
        """Permanently delete the workflow data with no confirmation."""
        self.path.unlink()

    @dropbox_permission_err_retry
    def remove_replaced_file(self) -> None:
        wk_data = self.load()
        if "replaced_file" in wk_data:
            Path(wk_data["replaced_file"]).unlink()
            self._pending["remove_replaced_file_record"] = True
            self.save()

    @dropbox_permission_err_retry
    def reinstate_replaced_file(self) -> None:
        wk_data = self.load()
        if "replaced_file" in wk_data:
            Path(wk_data["replaced_file"]).rename(self.path)

    def copy(self, path: PathLike = None) -> None:
        shutil.copy(self.path, path)

    def is_modified_on_disk(self) -> bool:
        if self._loaded:
            return get_md5_hash(self._load()) != get_md5_hash(self._loaded)
        else:
            # nothing to compare to
            return False



from __future__ import annotations
from contextlib import contextmanager
import copy
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
import shutil
import time
from typing import Dict, Iterator, List, Optional, Tuple, Type, Union
from warnings import warn

import numpy as np
import zarr

from hpcflow.sdk.typing import PathLike


from .json_like import ChildObjectSpec, JSONLike
from .parameters import InputSource
from .task import ElementSet, Task
from .utils import get_md5_hash, read_YAML, read_YAML_file
from .errors import (
    InvalidInputSourceTaskReference,
    WorkflowBatchUpdateFailedError,
)

from hpcflow.sdk.persistence.json import JSONPersistentStore
from hpcflow.sdk.persistence.zarr import ZarrPersistentStore

TS_NAME_FMT = r"%Y-%m-%d_%H%M%S"


@dataclass
class BatchUpdateData:
    """Class to store batch update instructions, ready for writing to the persistent
    workflow at the end of the batch update."""


@dataclass
class WorkflowTemplate(JSONLike):
    """Class to represent initial parametrisation of a workflow, with limited validation
    logic."""

    _child_objects = (
        ChildObjectSpec(
            name="tasks",
            class_name="Task",
            is_multiple=True,
            parent_ref="workflow_template",
        ),
    )

    name: str
    tasks: Optional[List[Task]] = field(default_factory=lambda: [])
    workflow: Optional[Workflow] = None

    def __post_init__(self):
        self._set_parent_refs()

    @classmethod
    def _from_data(cls, data: Dict) -> WorkflowTemplate:
        # use element_sets if not already:
        for task_idx, task_dat in enumerate(data["tasks"]):
            if "element_sets" not in task_dat:
                # add a single element set:
                elem_set = {}
                for chd_obj in ElementSet._child_objects:
                    if chd_obj.name in task_dat:
                        elem_set[chd_obj.name] = task_dat.pop(chd_obj.name)
                data["tasks"][task_idx]["element_sets"] = [elem_set]

        return cls.from_json_like(data, shared_data=cls.app.template_components)

    @classmethod
    def from_YAML_string(cls, string: str) -> WorkflowTemplate:
        return cls._from_data(read_YAML(string))

    @classmethod
    def from_YAML_file(cls, path: PathLike) -> WorkflowTemplate:
        return cls._from_data(read_YAML_file(path))

    def _add_empty_task(self, task: Task, new_index: int, insert_ID: int) -> None:
        """Called by `Workflow._add_empty_task`."""
        new_task_name = self.workflow._get_new_task_unique_name(task, new_index)

        task._insert_ID = insert_ID
        task._dir_name = f"task_{task.insert_ID}_{new_task_name}"
        task._element_sets = []  # element sets are added to the Task during add_elements
        task.workflow_template = self
        self.tasks.insert(new_index, task)


class Workflow:

    _app_attr = "app"

    _persistent_store_ext_lookup = {
        "json": ".json",
        "zarr": "",
    }

    _persistent_store_cls_lookup = {
        ".json": JSONPersistentStore,
        "": ZarrPersistentStore,
    }

    def __init__(self, path: PathLike) -> None:
        self.path = Path(path)

        # assigned on first access to corresponding properties:
        self._template = None
        self._template_components = None
        self._tasks = None

        self._store = self._get_store_class_from_ext(self.path)(self)

        self._in_batch_mode = False  # flag to track when processing batch updates

        # store indices of updates during batch update, so we can revert on failure:
        self._pending = self._get_empty_pending()

    def _get_empty_pending(self) -> Dict:
        return {
            "template_components": {k: [] for k in self.app._template_component_types},
            "tasks": [],
        }

    def _accept_pending(self) -> None:
        self._reset_pending()

    def _reset_pending(self) -> None:
        self._pending = self._get_empty_pending()

    def _reject_pending(self) -> None:
        """Revert pending changes to the in-memory representation of the workflow.

        This deletes new tasks and new template component data. Element additions to
        existing (non-pending) tasks are separately rejected/accepted by the
        WorkflowTask object.

        """
        for task_idx in self._pending["tasks"][::-1]:
            self.tasks._remove_object(task_idx)
            self.template.tasks.pop(task_idx)

        for comp_type, comp_indices in self._pending["template_components"].items():
            for comp_idx in comp_indices[::-1]:
                self.template_components[comp_type]._remove_object(comp_idx)

        self._reset_pending()

    @classmethod
    def _get_store_class_from_ext(cls, path: str) -> Type:
        return cls._persistent_store_cls_lookup[path.suffix.lower()]

    @property
    def store_format(self):
        # TODO: make this info cleaner to access
        for k, v in self._persistent_store_cls_lookup.items():
            if v == type(self._store):
                for k2, v2 in self._persistent_store_ext_lookup.items():
                    if v2 == k:
                        return k2

    @classmethod
    def from_template(
        cls,
        template: WorkflowTemplate,
        path: Optional[PathLike] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = "zarr",
    ) -> Workflow:
        wk = cls._write_empty_workflow(template, path, name, overwrite, store)
        with wk._store.cached_load():
            with wk.batch_update(is_workflow_creation=True):
                for task in template.tasks:
                    wk._add_task(task)
        return wk

    @classmethod
    def from_YAML_file(
        cls,
        YAML_path: PathLike,
        path: Optional[str] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
    ) -> Workflow:
        template = cls.app.WorkflowTemplate.from_YAML_file(YAML_path)
        return cls.from_template(template, path, name, overwrite)

    @classmethod
    def from_tasks(
        cls,
        name: str,
        tasks: List[Task],
        path: Optional[PathLike] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = "zarr",
    ) -> Workflow:
        raise NotImplementedError

    @contextmanager
    def batch_update(self, is_workflow_creation: bool = False) -> Iterator[None]:
        """A context manager that batches up structural changes to the workflow and
        commits them to disk all together when the context manager exits."""

        if self._in_batch_mode:
            yield
        else:
            try:
                self._in_batch_mode = True
                yield

            except Exception as err:

                print("batch update exception!")

                self._in_batch_mode = False
                self._store.reject_pending()

                for task in self.tasks:
                    task._reset_pending_elements()

                self._reject_pending()

                if is_workflow_creation:
                    # creation failed, so no need to keep the newly generated workflow:
                    self._store.delete_no_confirm()
                    self._store.reinstate_replaced_file()

                raise err

            else:

                if self._store.has_pending:

                    if self._store.is_modified_on_disk():
                        raise WorkflowBatchUpdateFailedError(
                            "Workflow modified on disk since it was loaded!"
                        )

                    for task in self.tasks:
                        task._accept_pending_elements()

                    self._store.remove_replaced_file()
                    self._store.commit_pending()
                    self._accept_pending()
                    self._in_batch_mode = False

    @classmethod
    def _write_empty_workflow(
        cls,
        template: WorkflowTemplate,
        path: Optional[PathLike] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = False,
        store: Optional[str] = "zarr",
    ) -> Workflow:

        timestamp = datetime.utcnow()

        path = Path(path or "").resolve()
        name = name or f"{template.name}_{timestamp.strftime(TS_NAME_FMT)}"
        ext = cls._persistent_store_ext_lookup[store.lower()]
        path = path.joinpath(name + ext)

        store_cls = cls._persistent_store_cls_lookup[ext]
        template_js, template_sh = template.to_json_like(exclude=["tasks"])
        template_js["tasks"] = []
        store_cls.write_empty_workflow(template_js, template_sh, path, overwrite)

        return cls(path)

    @property
    def num_tasks(self) -> int:
        return len(self.tasks)

    @property
    def num_added_tasks(self) -> int:
        with self._store.cached_load():
            return self._store.get_num_added_tasks()

    @property
    def num_elements(self) -> int:
        return sum(task.num_elements for task in self.tasks)

    @property
    def template_components(self) -> Dict:
        if self._template_components is None:
            with self._store.cached_load():
                tc_js = self._store.get_template_components()
            self._template_components = self.app.template_components_from_json_like(tc_js)
        return self._template_components

    @property
    def template(self) -> WorkflowTemplate:
        if self._template is None:
            with self._store.cached_load():
                temp_js = self._store.get_template()
                template = self.app.WorkflowTemplate.from_json_like(
                    temp_js, self.template_components
                )
                template.workflow = self
            self._template = template

        return self._template

    @property
    def tasks(self) -> WorkflowTaskList:
        if self._tasks is None:
            with self._store.cached_load():
                tasks_meta = self._store.get_all_tasks_metadata()
                wk_tasks = []
                for idx, i in enumerate(tasks_meta):
                    wk_task = self.app.WorkflowTask(
                        workflow=self,
                        template=self.template.tasks[idx],
                        index=idx,
                        num_elements=i["num_elements"],
                    )
                    wk_tasks.append(wk_task)
                self._tasks = self.app.WorkflowTaskList(wk_tasks)
        return self._tasks

    def elements(self) -> Iterator[Element]:
        for task in self.tasks:
            for element in task.elements:
                yield element

    def copy(self, path=None) -> Workflow:
        """Copy the workflow to a new path and return the copied workflow."""
        if path is None:
            path = self.path.parent / Path(self.path.stem + "_copy" + self.path.suffix)
        if path.exists():
            raise ValueError(f"Path already exists: {path}.")
        self._store.copy(path=path)
        return self.app.Workflow(path=path)

    def delete(self):
        self._store.delete()

    def _delete_no_confirm(self):
        self._store.delete_no_confirm()

    def rename(self, new_name: str):
        raise NotImplementedError

    def submit(self):
        raise NotImplementedError

    def add_submission(self, filter):
        raise NotImplementedError

    def get_task_unique_names(
        self, map_to_insert_ID: bool = False
    ) -> Union[List[str], Dict[str, int]]:
        """Return the unique names of all workflow tasks.

        Parameters
        ----------
        map_to_insert_ID : bool, optional
            If True, return a dict whose values are task insert IDs, otherwise return a
            list.

        """
        names = Task.get_task_unique_names(self.template.tasks)
        if map_to_insert_ID:
            insert_IDs = (i.insert_ID for i in self.template.tasks)
            return dict(zip(names, insert_IDs))
        else:
            return names

    def _get_new_task_unique_name(self, new_task: Task, new_index: int) -> str:

        task_templates = list(self.template.tasks)
        task_templates.insert(new_index, new_task)
        uniq_names = Task.get_task_unique_names(task_templates)

        return uniq_names[new_index]

    def _add_empty_task(
        self,
        task: Task,
        new_index: Optional[int] = None,
    ) -> WorkflowTask:

        if new_index is None:
            new_index = self.num_tasks

        insert_ID = self.num_added_tasks

        # make a copy with persistent schema inputs:
        task_c, _ = task.to_persistent(self, insert_ID)

        # add to the WorkflowTemplate:
        self.template._add_empty_task(task_c, new_index, insert_ID)

        # create and insert a new WorkflowTask:
        self.tasks.add_object(
            self.app.WorkflowTask.new_empty_task(self, task_c, new_index),
            index=new_index,
        )

        # update persistent store:
        task_js, temp_comps_js = task_c.to_json_like()
        self._store.add_template_components(temp_comps_js)
        self._store.add_empty_task(new_index, task_js)

        # update in-memory workflow template components:
        temp_comps = self.app.template_components_from_json_like(temp_comps_js)
        for comp_type, comps in temp_comps.items():
            for comp in comps:
                comp._set_hash()
                if comp not in self.template_components[comp_type]:
                    idx = self.template_components[comp_type].add_object(comp)
                    self._pending["template_components"][comp_type].append(idx)

        self._pending["tasks"].append(new_index)

        return self.tasks[new_index]

    def _add_task(self, task: Task, new_index: Optional[int] = None) -> None:
        new_wk_task = self._add_empty_task(task=task, new_index=new_index)
        new_wk_task._add_elements(element_sets=task.element_sets)

    def add_task(self, task: Task, new_index: Optional[int] = None) -> None:
        with self._store.cached_load():
            with self.batch_update():
                self._add_task(task, new_index=new_index)

    def add_task_after(self, task_ref):
        # TODO: find position of new task, then call add_task
        # TODO: add new downstream elements?
        pass

    def add_task_before(self, task_ref):
        # TODO: find position of new task, then call add_task
        # TODO: add new downstream elements?
        pass

    def _get_parameter_data(self, index: int) -> Any:
        return self._store.get_parameter_data(index)

    def _get_parameter_source(self, index: int) -> Dict:
        return self._store.get_parameter_source(index)

    def get_all_parameter_data(self) -> Dict[int, Any]:
        return self._store.get_all_parameter_data()

    def is_parameter_set(self, index: int) -> bool:
        return self._store.is_parameter_set(index)

    def check_parameters_exist(
        self, indices: Union[int, List[int]]
    ) -> Union[bool, List[bool]]:
        return self._store.check_parameters_exist(indices)

    def _add_unset_parameter_data(self, source: Dict) -> int:
        return self._store.add_unset_parameter_data(source)

    def _add_parameter_data(self, data, source: Dict) -> int:
        return self._store.add_parameter_data(data, source)

    def _resolve_input_source_task_reference(
        self, input_source: InputSource, new_task_name: str
    ) -> None:
        """Normalise the input source task reference and convert a source to a local type
        if required."""

        # TODO: test thoroughly!

        if isinstance(input_source.task_ref, str):
            if input_source.task_ref == new_task_name:
                if input_source.task_source_type is self.app.TaskSourceType.OUTPUT:
                    raise InvalidInputSourceTaskReference(
                        f"Input source {input_source.to_string()!r} cannot refer to the "
                        f"outputs of its own task!"
                    )
                else:
                    warn(
                        f"Changing input source {input_source.to_string()!r} to a local "
                        f"type, since the input source task reference refers to its own "
                        f"task."
                    )
                    # TODO: add an InputSource source_type setter to reset
                    # task_ref/source_type?
                    input_source.source_type = self.app.InputSourceType.LOCAL
                    input_source.task_ref = None
                    input_source.task_source_type = None
            else:
                try:
                    uniq_names_cur = self.get_task_unique_names(map_to_insert_ID=True)
                    input_source.task_ref = uniq_names_cur[input_source.task_ref]
                except KeyError:
                    raise InvalidInputSourceTaskReference(
                        f"Input source {input_source.to_string()!r} refers to a missing "
                        f"or inaccessible task: {input_source.task_ref!r}."
                    )

    def get_task_elements(self, task: Task, selection: slice) -> List[Element]:
        return [
            self.app.Element(task=task, **i)
            for i in self._store.get_task_elements(task.index, task.insert_ID, selection)
        ]

    def get_task_elements_islice(self, task: Task, selection: slice) -> Iterator[Element]:
        for i in self._store.get_task_elements_islice(
            task.index, task.insert_ID, selection
        ):
            yield self.app.Element(task=task, **i)

    def get_EARs_from_indices(
        self, indices: List[Tuple[int, int, int, int]]
    ) -> List[ElementActionRun]:
        """Return element action run objects from a list of four-tuples, representing the
        task insert ID, (local) element index, action index, and run index, respectively.
        """
        objs = []
        for src_i in indices:
            task = self.tasks.get(insert_ID=src_i[0])
            EAR_i = task.elements[src_i[1]].actions[src_i[2]].runs[src_i[3]]
            objs.append(EAR_i)
        return objs

    def get_elements_from_indices(self, indices: List[Tuple[int, int]]) -> List[Element]:
        """Return element objects from a list of two-tuples, representing the task insert
        ID, and (local) element index, respectively."""
        return [self.tasks.get(insert_ID=idx[0]).elements[idx[1]] for idx in indices]


@dataclass
class WorkflowBlueprint:
    """Pre-built workflow templates that are simpler to parametrise (e.g. fitting workflows)."""

    workflow_template: WorkflowTemplate



from __future__ import annotations
import copy
from dataclasses import dataclass, field
import enum
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np

from hpcflow.sdk.core.errors import (
    ValuesAlreadyPersistentError,
    MalformedParameterPathError,
    UnknownResourceSpecItemError,
    WorkflowParameterMissingError,
)
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.utils import check_valid_py_identifier
from hpcflow.sdk.core.zarr_io import ZarrEncodable, zarr_decode


Address = List[Union[int, float, str]]
Numeric = Union[int, float, np.number]


class ParameterValue:
    _typ = None

    def to_dict(self):
        if hasattr(self, "__dict__"):
            return dict(self.__dict__)
        elif hasattr(self, "__slots__"):
            return {k: getattr(self, k) for k in self.__slots__}


class ParameterPropagationMode(enum.Enum):

    IMPLICIT = 0
    EXPLICIT = 1
    NEVER = 2


@dataclass
class ParameterPath(JSONLike):

    path: Sequence[Union[str, int, float]]
    task: Optional[Union[TaskTemplate, TaskSchema]] = None  # default is "current" task


@dataclass
class Parameter(JSONLike):

    _validation_schema = "parameters_spec_schema.yaml"
    _child_objects = (
        ChildObjectSpec(
            name="typ",
            json_like_name="type",
        ),
    )

    typ: str
    is_file: bool = False
    sub_parameters: List[SubParameter] = field(default_factory=lambda: [])
    _value_class: Any = None
    _hash_value: Optional[str] = field(default=None, repr=False)

    def __repr__(self) -> str:

        is_file_str = ""
        if self.is_file:
            is_file_str = f", is_file={self.is_file!r}"

        sub_parameters_str = ""
        if self.sub_parameters:
            sub_parameters_str = f", sub_parameters={self.sub_parameters!r}"

        _value_class_str = ""
        if self._value_class is not None:
            _value_class_str = f", _value_class={self._value_class!r}"

        return (
            f"{self.__class__.__name__}("
            f"typ={self.typ!r}{is_file_str}{sub_parameters_str}{_value_class_str}"
            f")"
        )

    def __post_init__(self):
        self.typ = check_valid_py_identifier(self.typ)
        for i in ParameterValue.__subclasses__():
            if i._typ == self.typ:
                self._value_class = i

    def to_dict(self):
        dct = super().to_dict()
        del dct["_value_class"]
        return dct


@dataclass
class SubParameter:
    address: Address
    parameter: Parameter


@dataclass
class SchemaParameter(JSONLike):

    _app_attr = "app"

    _child_objects = (
        ChildObjectSpec(
            name="parameter",
            class_name="Parameter",
            shared_data_name="parameters",
            shared_data_primary_key="typ",
        ),
    )

    def __post_init__(self):
        self._validate()

    def _validate(self):
        if isinstance(self.parameter, str):
            self.parameter = self.app.Parameter(self.parameter)

    @property
    def name(self):
        return self.parameter.name

    @property
    def typ(self):
        return self.parameter.typ


@dataclass
class SchemaInput(SchemaParameter):
    """A Parameter as used within a particular schema, for which a default value may be
    applied."""

    _task_schema = None  # assigned by parent TaskSchema

    _child_objects = (
        ChildObjectSpec(
            name="parameter",
            class_name="Parameter",
            shared_data_name="parameters",
            shared_data_primary_key="typ",
        ),
        ChildObjectSpec(
            name="default_value",
            class_name="InputValue",
            parent_ref="_schema_input",
        ),
        ChildObjectSpec(
            name="propagation_mode",
            class_name="ParameterPropagationMode",
            is_enum=True,
        ),
    )

    parameter: Parameter
    default_value: Optional[InputValue] = None
    propagation_mode: ParameterPropagationMode = ParameterPropagationMode.IMPLICIT

    # can we define elements groups on local inputs as well, or should these be just for
    # elements from other tasks?
    group: Optional[str] = None
    where: Optional[ElementFilter] = None

    def __post_init__(self):
        super().__post_init__()
        self._set_parent_refs()

    def __repr__(self) -> str:

        default_str = ""
        if self.default_value is not None:
            default_str = f", default_value={self.default_value!r}"

        group_str = ""
        if self.group is not None:
            group_str = f", group={self.group!r}"

        where_str = ""
        if self.where is not None:
            where_str = f", group={self.where!r}"

        return (
            f"{self.__class__.__name__}("
            f"parameter={self.parameter.__class__.__name__}({self.parameter.typ!r}), "
            f"propagation_mode={self.propagation_mode.name!r}"
            f"{default_str}{group_str}{where_str}"
            f")"
        )

    def __deepcopy__(self, memo):
        kwargs = {
            "parameter": self.parameter,
            "default_value": self.default_value,
            "propagation_mode": self.propagation_mode,
            "group": self.group,
        }
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._task_schema = self._task_schema
        return obj

    @property
    def task_schema(self):
        return self._task_schema

    def _validate(self):
        super()._validate()
        if self.default_value is not None:
            if not isinstance(self.default_value, self.app.InputValue):
                self.default_value = self.app.InputValue(
                    parameter=self.parameter,
                    value=self.default_value,
                )
            if self.default_value.parameter != self.parameter:
                raise ValueError(
                    f"{self.__class__.__name__} `default_value` must be an `InputValue` for "
                    f"parameter: {self.parameter!r}, but specified `InputValue` parameter "
                    f"is: {self.default_value.parameter!r}."
                )

    @property
    def input_or_output(self):
        return "input"


@dataclass
class SchemaOutput(SchemaParameter):
    """A Parameter as outputted from particular task."""

    parameter: Parameter
    propagation_mode: ParameterPropagationMode = ParameterPropagationMode.IMPLICIT

    @property
    def input_or_output(self):
        return "output"

    def __repr__(self) -> str:

        return (
            f"{self.__class__.__name__}("
            f"parameter={self.parameter.__class__.__name__}({self.parameter.typ!r}), "
            f"propagation_mode={self.propagation_mode.name!r}"
            f")"
        )


@dataclass
class BuiltinSchemaParameter:
    # builtin inputs (resources,parameter_perturbations,method,implementation
    # builtin outputs (time, memory use, node/hostname etc)
    # - builtin parameters do not propagate to other tasks (since all tasks define the same
    #   builtin parameters).
    # - however, builtin parameters can be accessed if a downstream task schema specifically
    #   asks for them (e.g. for calculating/plotting a convergence test)
    pass


class ValueSequence(JSONLike):
    def __init__(
        self,
        path: str,
        nesting_order: int,
        values: List[Any],
        is_unused: bool = False,
    ):
        self.path = self._validate_parameter_path(path)
        self.nesting_order = nesting_order
        self.is_unused = is_unused  # TODO: what is this for; should it be in init?

        self._values = values

        self._values_group_idx = None
        self._workflow = None
        self._element_set = None  # assigned by parent ElementSet

        # assigned if this is an "inputs" sequence on validation of parent element set:
        self._parameter = None

        self._path_split = None  # assigned by property `path_split`

    def __repr__(self):
        vals_grp_idx = (
            f"values_group_idx={self._values_group_idx}, "
            if self._values_group_idx
            else ""
        )
        return (
            f"{self.__class__.__name__}("
            f"path={self.path!r}, "
            f"nesting_order={self.nesting_order}, "
            f"{vals_grp_idx}"
            f"values={self.values}"
            f")"
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if self.to_dict() == other.to_dict():
            return True
        return False

    def __deepcopy__(self, memo):
        kwargs = self.to_dict()
        kwargs["values"] = kwargs.pop("_values")
        _values_group_idx = kwargs.pop("_values_group_idx")
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._values_group_idx = _values_group_idx
        obj._workflow = self._workflow
        obj._element_set = self._element_set
        obj._path_split = self._path_split
        obj._parameter = self._parameter
        return obj

    @property
    def parameter(self):
        return self._parameter

    @property
    def path_split(self):
        if self._path_split is None:
            self._path_split = self.path.split(".")
        return self._path_split

    @property
    def path_type(self):
        return self.path_split[0]

    @property
    def input_type(self):
        if self.path_type == "inputs":
            return self.path_split[1]

    @property
    def input_path(self):
        if self.path_type == "inputs":
            return ".".join(self.path_split[2:]) or None

    @property
    def resource_scope(self):
        if self.path_type == "resources":
            return self.path_split[1]

    @property
    def is_sub_value(self):
        """True if the values are for a sub part of the parameter."""
        return True if self.input_path else False

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""

        _values_group_idx = json_like.pop("_values_group_idx", None)
        if "_values" in json_like:
            json_like["values"] = json_like.pop("_values")

        obj = cls(**json_like)
        obj._values_group_idx = _values_group_idx
        return obj

    def _validate_parameter_path(self, path):

        if not isinstance(path, str):
            raise MalformedParameterPathError(
                f"`path` must be a string, but given path has type {type(path)} with value "
                f"{path!r}."
            )
        path = path.lower()
        path_split = path.split(".")
        if not path_split[0] in ("inputs", "resources"):
            raise MalformedParameterPathError(
                f'`path` must start with "inputs", "outputs", or "resources", but given path '
                f"is: {path!r}."
            )
        if path_split[0] == "resources":
            try:
                self.app.ActionScope.from_json_like(path_split[1])
            except Exception as err:
                raise MalformedParameterPathError(
                    f"Cannot parse a resource action scope from the second component of the "
                    f"path: {path!r}. Exception was: {err}."
                ) from None

            if len(path_split) > 2:
                path_split_2 = path_split[2]
                allowed = ResourceSpec.ALLOWED_PARAMETERS
                if path_split_2 not in allowed:
                    allowed_keys_str = ", ".join(f'"{i}"' for i in allowed)
                    raise UnknownResourceSpecItemError(
                        f"Resource item name {path_split_2!r} is unknown. Allowed "
                        f"resource item names are: {allowed_keys_str}."
                    )

        return path

    def to_dict(self):
        out = super().to_dict()
        del out["_parameter"]
        del out["_path_split"]
        if "_workflow" in out:
            del out["_workflow"]
        return out

    @property
    def normalised_path(self):
        return self.path

    @property
    def normalised_inputs_path(self):
        """Return the normalised path without the "inputs" prefix, if the sequence is an
        inputs sequence, else return None."""

        if self.parameter:
            return ".".join(self.path_split[1:])

    def make_persistent(self, workflow, source):
        """Save value to a persistent workflow."""

        if self._values_group_idx is not None:
            is_new = False
            data_ref = self._values_group_idx
            if not all(workflow.check_parameters_exist(data_ref)):
                raise RuntimeError(
                    f"{self.__class__.__name__} has a parameter group index "
                    f"({data_ref}), but does not exist in the workflow."
                )
            # TODO: log if already persistent.

        else:
            data_ref = []
            for idx, i in enumerate(self._values):
                source = copy.deepcopy(source)
                source["sequence_idx"] = idx
                pg_idx_i = workflow._add_parameter_data(i, source=source)
                data_ref.append(pg_idx_i)

            is_new = True
            self._values_group_idx = data_ref
            self._workflow = workflow
            self._values = None

        return (self.normalised_path, data_ref, is_new)

    @property
    def workflow(self):
        if self._workflow:
            return self._workflow
        elif self._element_set:
            return self._element_set.task_template.workflow_template.workflow

    @property
    def values(self):
        if self._values_group_idx is not None:
            vals = []
            for pg_idx_i in self._values_group_idx:
                val = self.workflow._get_parameter_data(pg_idx_i)
                if self.parameter._value_class:
                    val = self.parameter._value_class(**val)
                vals.append(val)
            return vals
        else:
            return self._values

    @classmethod
    def from_linear_space(cls, start, stop, nesting_order, num=50, path=None, **kwargs):
        values = list(np.linspace(start, stop, num=num, **kwargs))
        # TODO: save persistently as an array?
        return cls(values=values, path=path, nesting_order=nesting_order)

    @classmethod
    def from_range(cls, start, stop, nesting_order, step=1, path=None):
        if isinstance(step, int):
            return cls(
                values=list(np.arange(start, stop, step)),
                path=path,
                nesting_order=nesting_order,
            )
        else:
            # Use linspace for non-integer step, as recommended by Numpy:
            return cls.from_linear_space(
                start,
                stop,
                num=int((stop - start) / step),
                address=path,
                endpoint=False,
                nesting_order=nesting_order,
            )


@dataclass
class AbstractInputValue(JSONLike):
    """Class to represent all sequence-able inputs to a task."""

    _workflow = None

    def __repr__(self):
        try:
            value_str = f", value={self.value}"
        except WorkflowParameterMissingError:
            value_str = ""

        return (
            f"{self.__class__.__name__}("
            f"_value_group_idx={self._value_group_idx}"
            f"{value_str}"
            f")"
        )

    def to_dict(self):
        out = super().to_dict()
        if "_workflow" in out:
            del out["_workflow"]
        return out

    def make_persistent(self, workflow, source) -> Dict:
        """Save value to a persistent workflow.

        Parameters
        ----------
        workflow : Workflow

        Returns
        -------
        (str, list of int)
            String is the data path for this task input and single item integer list
            contains the index of the parameter data Zarr group where the data is
            stored.
        """

        if self._value_group_idx is not None:
            data_ref = self._value_group_idx
            is_new = False
            if not workflow.check_parameters_exist(data_ref):
                raise RuntimeError(
                    f"{self.__class__.__name__} has a data reference "
                    f"({data_ref}), but does not exist in the workflow."
                )
            # TODO: log if already persistent.
        else:
            data_ref = workflow._add_parameter_data(self._value, source=source)
            self._value_group_idx = data_ref
            is_new = True
            self._value = None

        return (self.normalised_path, [data_ref], is_new)

    @property
    def workflow(self):
        if self._workflow:
            return self._workflow
        elif self._element_set:
            return self._element_set.task_template.workflow_template.workflow
        elif self._schema_input:
            return self._schema_input.task_schema.task_template.workflow_template.workflow

    @property
    def value(self):
        if self._value_group_idx is not None:
            val = self.workflow._get_parameter_data(self._value_group_idx)
            if self.parameter._value_class:
                val = self.parameter._value_class(**val)
        else:
            val = self._value

        return val


@dataclass
class ValuePerturbation(AbstractInputValue):
    name: str
    path: Optional[Sequence[Union[str, int, float]]] = None
    multiplicative_factor: Optional[Numeric] = 1
    additive_factor: Optional[Numeric] = 0

    @classmethod
    def from_spec(cls, spec):
        return cls(**spec)


class InputValue(AbstractInputValue):

    _child_objects = (
        ChildObjectSpec(
            name="parameter",
            class_name="Parameter",
            shared_data_primary_key="typ",
            shared_data_name="parameters",
        ),
    )

    def __init__(
        self,
        parameter: Union[Parameter, str],
        value: Optional[Any] = None,
        path: Optional[str] = None,
    ):
        if isinstance(parameter, str):
            parameter = self.app.parameters.get(parameter)
        elif isinstance(parameter, SchemaInput):
            parameter = parameter.parameter

        self.parameter = parameter
        self.path = (path.strip(".") if path else None) or None
        self._value = value

        self._value_group_idx = None  # assigned by method make_persistent
        self._element_set = None  # assigned by parent ElementSet (if belonging)

        # assigned by parent SchemaInput (if this object is a default value of a
        # SchemaInput):
        self._schema_input = None

    def __deepcopy__(self, memo):
        kwargs = self.to_dict()
        _value = kwargs.pop("_value")
        _value_group_idx = kwargs.pop("_value_group_idx")
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._value = _value
        obj._value_group_idx = _value_group_idx
        obj._element_set = self._element_set
        obj._schema_input = self._schema_input
        return obj

    def __repr__(self):

        val_grp_idx = ""
        if self._value_group_idx is not None:
            val_grp_idx = f", value_group_idx={self._value_group_idx}"

        path_str = ""
        if self.path is not None:
            path_str = f", path={self.path!r}"

        try:
            value_str = f", value={self.value}"
        except WorkflowParameterMissingError:
            value_str = ""

        return (
            f"{self.__class__.__name__}("
            f"parameter={self.parameter.typ!r}"
            f"{value_str}"
            f"{path_str}"
            f"{val_grp_idx}"
            f")"
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if self.to_dict() == other.to_dict():
            return True
        return False

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""

        _value_group_idx = json_like.pop("_value_group_idx", None)
        if "_value" in json_like:
            json_like["value"] = json_like.pop("_value")

        obj = cls(**json_like)
        obj._value_group_idx = _value_group_idx

        return obj

    @property
    def normalised_inputs_path(self):
        return f"{self.parameter.typ}" f"{f'.{self.path}' if self.path else ''}"

    @property
    def normalised_path(self):
        return f"inputs.{self.normalised_inputs_path}"

    @classmethod
    def from_json_like(cls, json_like, shared_data=None):

        if "path" not in json_like:
            param_spec = json_like["parameter"].split(".")
            json_like["parameter"] = param_spec[0]
            json_like["path"] = ".".join(param_spec[1:])

        obj = super().from_json_like(json_like, shared_data)

        return obj

    @property
    def is_sub_value(self):
        """True if the value is for a sub part of the parameter (i.e. if `path` is set).
        Sub-values are not added to the base parameter data, but are interpreted as
        single-value sequences."""
        return True if self.path else False


class ResourceSpec(JSONLike):

    ALLOWED_PARAMETERS = {
        "scratch",
        "num_cores",
    }

    _resource_list = None

    _child_objects = (
        ChildObjectSpec(
            name="scope",
            class_name="ActionScope",
        ),
    )

    def __init__(self, scope=None, scratch=None, num_cores=None):

        self.scope = scope or self.app.ActionScope.any()

        # user-specified resource parameters:
        self._scratch = scratch
        self._num_cores = num_cores

        # assigned by `make_persistent`
        self._workflow = None
        self._value_group_idx = None

    def __deepcopy__(self, memo):
        kwargs = copy.deepcopy(self.to_dict(), memo)
        _value_group_idx = kwargs.pop("value_group_idx")
        obj = self.__class__(**kwargs)
        obj._value_group_idx = _value_group_idx
        obj._resource_list = self._resource_list
        return obj

    def __repr__(self):
        param_strs = ""
        for i in self.ALLOWED_PARAMETERS:
            i_str = ""
            try:
                i_val = getattr(self, i)
            except WorkflowParameterMissingError:
                pass
            else:
                if i_val is not None:
                    i_str = f", {i}={i_val}"

            param_strs += i_str

        return f"{self.__class__.__name__}(scope={self.scope}{param_strs})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if self.to_dict() == other.to_dict():
            return True
        return False

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""

        _value_group_idx = json_like.pop("value_group_idx", None)
        try:
            obj = cls(**json_like)
        except TypeError:
            given_keys = set(k for k in json_like.keys() if k != "scope")
            bad_keys = given_keys - cls.ALLOWED_PARAMETERS
            bad_keys_str = ", ".join(f'"{i}"' for i in bad_keys)
            allowed_keys_str = ", ".join(f'"{i}"' for i in cls.ALLOWED_PARAMETERS)
            raise UnknownResourceSpecItemError(
                f"The following resource item names are unknown: {bad_keys_str}. Allowed "
                f"resource item names are: {allowed_keys_str}."
            )
        obj._value_group_idx = _value_group_idx

        return obj

    @property
    def normalised_resources_path(self):
        return self.scope.to_string()

    @property
    def normalised_path(self):
        return f"resources.{self.normalised_resources_path}"

    def to_dict(self):
        out = super().to_dict()
        if "_workflow" in out:
            del out["_workflow"]

        if self._value_group_idx is not None:
            # only store pointer to persistent data:
            out = {k: v for k, v in out.items() if k in ["_value_group_idx", "scope"]}

        out = {k.lstrip("_"): v for k, v in out.items()}
        return out

    def _get_members(self):
        out = self.to_dict()
        del out["scope"]
        del out["value_group_idx"]
        return out

    def make_persistent(self, workflow, source) -> Dict:
        """Save to a persistent workflow.

        Parameters
        ----------
        workflow : Workflow

        Returns
        -------

        (str, list of int)
            String is the data path for this task input and integer list
            contains the indices of the parameter data Zarr groups where the data is
            stored.
        """
        if self._value_group_idx is not None:
            data_ref = self._value_group_idx
            is_new = False
            if not workflow.check_parameters_exist(data_ref):
                raise RuntimeError(
                    f"{self.__class__.__name__} has a parameter group index "
                    f"({data_ref}), but does not exist in the workflow."
                )
            # TODO: log if already persistent.
        else:
            data_ref = workflow._add_parameter_data(self._get_members(), source=source)
            is_new = True
            self._value_group_idx = data_ref
            self._workflow = workflow
            self._num_cores = None
            self._scratch = None

        return (self.normalised_path, [data_ref], is_new)

    def _get_value(self, value_name=None):
        if self._value_group_idx is not None:
            val = self.workflow._get_parameter_data(self._value_group_idx)
        else:
            val = self._get_members()
        if value_name:
            val = val.get(value_name)

        return val

    @property
    def scratch(self):
        return self._get_value("scratch")

    @property
    def num_cores(self):
        return self._get_value("num_cores")

    @property
    def workflow(self):
        if self._workflow:
            return self._workflow
        elif self.element_set:
            return self.element_set.task_template.workflow_template.workflow

    @property
    def element_set(self):
        return self._resource_list.element_set


class InputSourceType(enum.Enum):

    IMPORT = 0
    LOCAL = 1
    DEFAULT = 2
    TASK = 3


class TaskSourceType(enum.Enum):
    INPUT = 0
    OUTPUT = 1
    ANY = 2


class InputSource(JSONLike):

    _child_objects = (
        ChildObjectSpec(
            name="source_type",
            json_like_name="type",
            class_name="InputSourceType",
            is_enum=True,
        ),
    )

    def __init__(
        self,
        source_type,
        import_ref=None,
        task_ref=None,
        task_source_type=None,
        elements=None,
        path=None,
        where=None,
    ):

        self.source_type = self._validate_source_type(source_type)
        self.import_ref = import_ref
        self.task_ref = task_ref
        self.task_source_type = self._validate_task_source_type(task_source_type)
        self.elements = elements
        self.where = where
        self.path = path

        if self.source_type is InputSourceType.TASK:
            if self.task_ref is None:
                raise ValueError(f"Must specify `task_ref` if `source_type` is TASK.")
            if self.task_source_type is None:
                self.task_source_type = TaskSourceType.OUTPUT

        if self.source_type is InputSourceType.IMPORT and self.import_ref is None:
            raise ValueError(f"Must specify `import_ref` if `source_type` is IMPORT.")

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        elif (
            self.source_type == other.source_type
            and self.import_ref == other.import_ref
            and self.task_ref == other.task_ref
            and self.task_source_type == other.task_source_type
            and self.elements == other.elements
            and self.where == other.where
            and self.path == other.path
        ):
            return True
        else:
            return False

    def __repr__(self) -> str:
        cls_method_name = self.source_type.name.lower()

        if self.source_type is InputSourceType.IMPORT:
            cls_method_name += "_"
            args = f"import_ref={self.import_ref}"

        elif self.source_type is InputSourceType.TASK:
            args = (
                f"task_ref={self.task_ref}, "
                f"task_source_type={self.task_source_type.name.lower()!r}"
            )
            if self.elements:
                args += f", elements={self.elements}"
        else:
            args = ""

        out = f"{self.__class__.__name__}.{cls_method_name}({args})"

        return out

    def get_task(self, workflow):
        """If source_type is task, then return the referenced task from the given
        workflow."""
        if self.source_type is InputSourceType.TASK:
            for task in workflow.tasks:
                if task.insert_ID == self.task_ref:
                    return task

    def is_in(self, other_input_sources):
        """Check if this input source is in a list of other input sources, without
        considering the `elements` attribute."""

        for other in other_input_sources:
            if (
                self.source_type == other.source_type
                and self.import_ref == other.import_ref
                and self.task_ref == other.task_ref
                and self.task_source_type == other.task_source_type
                and self.where == other.where
                and self.path == other.path
            ):
                return True
        return False

    def to_string(self):
        out = [self.source_type.name.lower()]
        if self.source_type is InputSourceType.TASK:
            out += [str(self.task_ref), self.task_source_type.name.lower()]
            if self.elements:
                out += ["[" + ",".join(f"{i}" for i in self.elements) + "]"]
        elif self.source_type is InputSourceType.IMPORT:
            out += [str(self.import_ref)]
        return ".".join(out)

    @staticmethod
    def _validate_source_type(src_type):
        if src_type is None:
            return None
        if isinstance(src_type, InputSourceType):
            return src_type
        try:
            src_type = getattr(InputSourceType, src_type.upper())
        except AttributeError:
            raise ValueError(
                f"InputSource `source_type` specified as {src_type!r}, but "
                f"must be one of: {[i.name for i in InputSourceType]!r}."
            )
        return src_type

    @staticmethod
    def _validate_task_source_type(task_src_type):
        if task_src_type is None:
            return None
        if isinstance(task_src_type, TaskSourceType):
            return task_src_type
        try:
            task_source_type = getattr(TaskSourceType, task_src_type.upper())
        except AttributeError:
            raise ValueError(
                f"InputSource `task_source_type` specified as {task_src_type!r}, but "
                f"must be one of: {[i.name for i in TaskSourceType]!r}."
            )
        return task_source_type

    @classmethod
    def from_string(cls, str_defn):
        return cls(**cls._parse_from_string(str_defn))

    @classmethod
    def _parse_from_string(cls, str_defn):
        """Parse a dot-delimited string definition of an InputSource.

        Examples:
            - task.[task_ref].input
            - task.[task_ref].output
            - local
            - default
            - import.[import_ref]

        """
        parts = str_defn.lower().split(".")
        source_type = cls._validate_source_type(parts[0])
        task_ref = None
        task_source_type = None
        import_ref = None
        if (
            (
                source_type in (InputSourceType.LOCAL, InputSourceType.DEFAULT)
                and len(parts) > 1
            )
            or (source_type is InputSourceType.TASK and len(parts) > 3)
            or (source_type is InputSourceType.IMPORT and len(parts) > 2)
        ):
            raise ValueError(f"InputSource string not understood: {str_defn!r}.")

        if source_type is InputSourceType.TASK:
            task_ref = parts[1]
            try:
                task_ref = int(task_ref)
            except ValueError:
                pass
            try:
                task_source_type_str = parts[2]
            except IndexError:
                task_source_type_str = TaskSourceType.OUTPUT
            task_source_type = cls._validate_task_source_type(task_source_type_str)
        elif source_type is InputSourceType.IMPORT:
            import_ref = parts[1]
            try:
                import_ref = int(import_ref)
            except ValueError:
                pass

        return {
            "source_type": source_type,
            "import_ref": import_ref,
            "task_ref": task_ref,
            "task_source_type": task_source_type,
        }

    @classmethod
    def from_json_like(cls, json_like, shared_data=None):
        if isinstance(json_like, str):
            json_like = cls._parse_from_string(json_like)
        return super().from_json_like(json_like, shared_data)

    @classmethod
    def import_(cls, import_ref):
        return cls(source_type=InputSourceType.IMPORT, import_ref=import_ref)

    @classmethod
    def local(cls):
        return cls(source_type=InputSourceType.LOCAL)

    @classmethod
    def default(cls):
        return cls(source_type=InputSourceType.DEFAULT)

    @classmethod
    def task(cls, task_ref, task_source_type=None, elements=None):
        if not task_source_type:
            task_source_type = TaskSourceType.OUTPUT
        return cls(
            source_type=InputSourceType.TASK,
            task_ref=task_ref,
            task_source_type=cls._validate_task_source_type(task_source_type),
            elements=elements,
        )



"""Module defining a function that returns the click CLI group for manipulating the app
configuration."""

import logging
import warnings
from functools import wraps
from contextlib import contextmanager

import click
from colorama import init as colorama_init
from termcolor import colored

from .errors import ConfigError

logger = logging.getLogger(__name__)

colorama_init(autoreset=True)


def custom_warning_formatter(message, category, filename, lineno, file=None, line=None):
    """Simple warning formatter that shows just the warning type and the message. We use
    this in the CLI, to avoid producing distracting output."""
    return f"{colored(category.__name__, 'yellow')}: {message}\n"


@contextmanager
def warning_formatter(func=custom_warning_formatter):
    """Context manager for modifying the warnings formatter.

    Parameters
    ----------
    func : function to set as the `warnings.formatwarning` function.

    """
    existing_func = warnings.formatwarning
    try:
        warnings.formatwarning = func
        yield
    finally:
        warnings.formatwarning = existing_func


def CLI_exception_wrapper_gen(*exception_cls):
    """Decorator factory"""

    def CLI_exception_wrapper(func):
        """Decorator

        Parameters
        ----------
        func
            Function that return a truthy value if the ???
        """

        @wraps(func)
        @click.pass_context
        def wrapper(ctx, *args, **kwargs):
            try:
                with warning_formatter():
                    out = func(*args, **kwargs)
                if out is not None:
                    click.echo(f"{colored('✔ Config file updated.', 'green')}")
                return out
            except exception_cls as err:
                click.echo(f"{colored(err.__class__.__name__, 'red')}: {err}")
                ctx.exit(1)

        return wrapper

    return CLI_exception_wrapper


def get_config_CLI(app):
    """Generate the configuration CLI for the app."""

    def show_all_config(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        print(app.config.to_string(exclude=["config_file_contents"]))
        ctx.exit()

    def show_config_file(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        print(app.config.config_file_contents)
        ctx.exit()

    @click.group()
    @click.option("--invocation", default="_")
    def config(invocation):
        """Configuration sub-command for getting and setting data in the configuration
        file(s)."""

    @config.command("list")
    def config_list():
        """Show a list of all configurable keys."""
        click.echo("\n".join(app.config.get_configurable()))

    @config.command()
    @click.argument("name")
    @click.option(
        "--all",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        help="Show all configuration items.",
        callback=show_all_config,
    )
    @click.option(
        "--file",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        help="Show the contents of the configuration file.",
        callback=show_config_file,
    )
    @CLI_exception_wrapper_gen(ConfigError)
    def get(name):
        """Show the value of the specified configuration item."""
        val = app.config.get(name)
        if isinstance(val, list):
            val = "\n".join(str(i) for i in val)
        click.echo(val)

    @config.command()
    @click.argument("name")
    @click.argument("value")
    @click.option(
        "--json",
        "is_json",
        is_flag=True,
        default=False,
        help="Interpret VALUE as a JSON string.",
    )
    @CLI_exception_wrapper_gen(ConfigError)
    def set(name, value, is_json):
        """Set and save the value of the specified configuration item."""
        app.config.set(name, value, is_json)
        app.config.save()

    @config.command()
    @click.argument("name")
    @CLI_exception_wrapper_gen(ConfigError)
    def unset(name):
        """Unset and save the value of the specified configuration item."""
        app.config.unset(name)
        app.config.save()

    @config.command()
    @click.argument("name")
    @click.argument("value")
    @CLI_exception_wrapper_gen(ConfigError)
    def append(name, value):
        """Append a new value to the specified configuration item."""
        app.config.append(name, value)
        app.config.save()

    @config.command()
    @click.argument("name")
    @click.argument("value")
    @CLI_exception_wrapper_gen(ConfigError)
    def prepend(name, value):
        """Append a new value to the specified configuration item."""
        app.config.prepend(name, value)
        app.config.save()

    @config.command(context_settings={"ignore_unknown_options": True})
    @click.argument("name")
    @click.argument("index", type=click.types.INT)
    @CLI_exception_wrapper_gen(ConfigError)
    def pop(name, index):
        """Remove a value from a list-like configuration item."""
        app.config.pop(name, index)
        app.config.save()

    @config.command()
    def load_data_files():
        """Check we can load the data files (e.g. task schema files) as specified in the
        configuration."""
        app.load_data_files()

    return config
